from typing import Any

import hypothesis.strategies as hs
from astroid import nodes
from hypothesis import HealthCheck, assume, given, settings
from pytest import skip

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


@given(cs.boolop_node(value=cs.const_node(hs.integers())))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_homogeneous_binary_boolop(node):
    """Test type setting of binary BoolOp node(s) representing expression with homogeneous operands."""
    module, _ = cs._parse_text(node)
    boolop_node = list(module.nodes_of_class(nodes.BoolOp))[0]
    assert boolop_node.inf_type.getValue() == boolop_node.values[0].inf_type.getValue()


@given(cs.boolop_node())
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_heterogeneous_binary_boolop(node):
    """Test type setting of binary BoolOp node(s) representing expression with heterogeneous operands."""
    skip(
        "Currently fails due to typechecking for inheritance. "
        "Need to figure out if this is expected behavior."
    )
    assume(type(node.values[0].value) != type(node.values[1].value))
    module, _ = cs._parse_text(node)
    boolop_node = list(module.nodes_of_class(nodes.BoolOp))[0]
    assert boolop_node.inf_type.getValue() == Any



from collections import OrderedDict
from math import ceil, sqrt
from statistics import median


def _individual_calc(error_msgs, style_msgs):
    """
    Analyses the given lists of error and style Message objects error_msgs and
    style_msgs for an individual.

    @param List[Message] error_msgs: all of this individual's code errors
    @param List[Message] style_msgs: all of this individual's style issues
    @rtype: List[Tuple[str, List]]
    """
    # {msg.symbol + "(" + msg.object + ")": count}
    all_msgs = error_msgs + style_msgs
    totals = [
        ("Error and Style Messages", len(all_msgs)),
        ("Error Messages", len(error_msgs)),
        ("Style Messages", len(style_msgs)),
    ]

    all_num = list(zip(*_calc_helper(all_msgs)))
    error_num = list(zip(*_calc_helper(error_msgs)))
    style_num = list(zip(*_calc_helper(style_msgs)))

    return [
        ("Totals", totals),
        ("Most Frequent Messages", all_num),
        ("Most Frequent Errors", error_num),
        ("Most Frequent Styles", style_num),
    ]


def summary(all_msgs):
    """
    Provides a summary of each individual's errors from all_msgs and provide an
    overall summary of the course's performance (if applicable).
    Called by pyta_statistics.

    @param OrderedDict[str, Tuple[List[Message], List[Message]]] all_msgs:
        the tuple of code and error messages for each student's files
    @rtype: Tuple[OrderedDict[str, List]]]
    """
    num_stu = len(all_msgs)

    # If directory was for student, not course, return empty summary stats list.
    if num_stu == 1:
        student, stats = all_msgs.popitem()
        return OrderedDict([(student, _individual_calc(*stats))]), OrderedDict()

    indiv_stats = OrderedDict()
    code_errors = []
    style_errors = []
    stu_errors = []

    for student, (errors, styles) in all_msgs.items():
        # in the form {std1': (<error>, <style>), 'std2': (<error>, <style>), }
        indiv_stats[student] = _individual_calc(errors, styles)
        stu_errors.append(len(errors) + len(styles))

        # To find Most Frequent Errors (aggregate)
        code_errors.extend(errors)
        style_errors.extend(styles)

    error_num = _frequent_messages(_message_counter(code_errors))
    style_num = _frequent_messages(_message_counter(style_errors))
    both_num = _frequent_messages(_message_counter(code_errors + style_errors))

    avg_errors = round((len(code_errors) + len(style_errors)) / num_stu, 2)
    # Calculating the Five Number Summary for all errors (per student)
    stu_errors.sort(reverse=True)

    med = median(stu_errors)

    # improving quartiles without numpy, as per StackOverflow.com/a/15589202
    q3 = stu_errors[ceil(0.25 * len(stu_errors)) - 1]
    q1 = stu_errors[ceil(0.75 * len(stu_errors)) - 1]

    # Calculating the Standard Deviation
    sum_dev = 0
    for error_dev in stu_errors:
        sum_dev += (error_dev - avg_errors) ** 2

    std_dev = sqrt(sum_dev / len(stu_errors))

    sum_stats = [
        ("Most Frequent Code Errors", error_num),
        ("Average Code Errors Per Student", round(len(code_errors) / num_stu, 2)),
        ("Most Frequent Style Errors", style_num),
        ("Average Style Errors Per Student", round(len(style_errors) / num_stu, 2)),
        ("Most Frequent Errors of Either Type", both_num),
        ("Average Errors of Either Type Per Student", avg_errors),
        (
            "Five Number Summary of Errors Per Student",
            [
                ("Most Errors", stu_errors[0]),
                ("Upper Quartile (Q3)", q3),
                ("Median", med),
                ("Lower Quartile (Q1)", q1),
                ("Least Errors", stu_errors[-1]),
            ],
        ),
        ("Standard Deviation", round(std_dev, 2)),
    ]

    return indiv_stats, sum_stats


def _calc_helper(msgs):
    """Returns frequent messages in numbers and in percentages.

    @param List[Message] msgs: Message objects for all errors found by linters
    @rtype: List[List]
    """
    # get dict of values {id:int, id2:int}
    msgs_dict = _message_counter(msgs)
    # sort into list of tuple, highest on top
    freq_nums = _frequent_messages(msgs_dict)
    total_msgs = len(msgs)
    # divide each value by total and round to two places
    for message in msgs_dict:
        msgs_dict[message] = round((msgs_dict[message] / total_msgs * 100), 2)
    perc_nums = _frequent_messages(msgs_dict)
    return [freq_nums, perc_nums]


def _message_counter(msgs):
    """Return the number of errors for every type of error in msgs.

    @param List[Message] msgs: the messages to count
    @rtype: Dict[str, int]
    """
    msgs_dict = {}

    for msg in msgs:
        key = "{} ({})".format(msg.msg_id, msg.symbol)
        if key not in msgs_dict:
            msgs_dict[key] = sum(1 for m in msgs if m.msg_id == msg.msg_id)
    return msgs_dict


def _frequent_messages(comp_dict, top=5):
    """
    Sort the errors in comp_dict from the most frequent to least frequent in a
    list.
    Return <top> most frequently occurring errors.

    @type comp_dict: Dict[str, number]
    @type top: int
    @rtype: List[Tuple[str, number]]
    """
    # get key-value pair in a list
    most_frequently = list(comp_dict.items())
    # sort the list according to the value, from larger number to lower number
    most_frequently.sort(key=lambda p: p[1], reverse=True)
    # So the name of the error first and then the number of its occurrence.
    # return the top whatever number
    return most_frequently[:top]



from typing import Any, List

from astroid import nodes
from hypothesis import HealthCheck, assume, given, settings
from pytest import skip

from python_ta.typecheck.base import TypeFail

from .. import custom_hypothesis_support as cs
from ..custom_hypothesis_support import lookup_type

settings.load_profile("pyta")


@given(cs.simple_homogeneous_list_node(min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_homogeneous_lists(lst):
    """Test List nodes representing a list of values of the same primitive type."""
    module, _ = cs._parse_text(lst)
    list_node = list(module.nodes_of_class(nodes.List))[0]
    if len(list_node.elts) == 0:
        assert list_node.inf_type.getValue() == List[Any]
    else:
        cs._verify_type_setting(module, nodes.List, List[type(lst.elts[0].value)])


@given(cs.list_node(min_size=2))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_random_lists(lst):
    """Test List nodes representing a list of values of different primitive types."""
    assume(not isinstance(lst.elts[0].value, type(lst.elts[1].value)))
    assume(not isinstance(lst.elts[1].value, type(lst.elts[0].value)))
    val_types = [type(val.value) for val in lst.elts]
    if int in val_types:
        assume(bool not in val_types)
    if bool in val_types:
        assume(int not in val_types)
    module, _ = cs._parse_text(lst)
    cs._verify_type_setting(module, nodes.List, List[Any])


def test_empty_list_reassign():
    src = """
    x = []
    x = [1]
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    x = [ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)][0]
    assert ti.type_constraints.resolve(x).getValue() == List[int]


def test_empty_list_reassign_invalid():
    src = """
    x = [1]
    x = ['abc']
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    assgn_node = list(ast_mod.nodes_of_class(nodes.Assign))[1]
    assert isinstance(assgn_node.inf_type, TypeFail)


def test_empty_list_reassign_twice():
    skip("This test requires special treatment of Any in generics")
    src = """
    x = [] # List[~T1]
    x = [1] # List[int]
    x = [1, 'abc'] # List[Any] and not List[int]
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    x = [ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)][0]
    assert ti.type_constraints.resolve(x).getValue() == List[Any]


def test_empty_list_append():
    src = """
    x = []
    x.append(1)
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    x = [ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)][0]
    assert ti.type_constraints.resolve(x).getValue() == List[int]


def test_empty_list_append_invalid():
    src = """
    x = []
    x.append(1)
    x.append('abc')
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    call_node = list(ast_mod.nodes_of_class(nodes.Call))[1]
    assert isinstance(call_node.inf_type, TypeFail)


def test_list_append():
    src = """
    x = [1,2,3]
    x.append(4)
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    x = [ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)][0]
    assert ti.type_constraints.resolve(x).getValue() == List[int]


def test_list_append_list_typevar():
    src = """
    def f(x):
        lst = [x]
        lst.append(lst)
        return lst
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    return_node = next(ast_mod.nodes_of_class(nodes.Return))
    functiondef_node = next(ast_mod.nodes_of_class(nodes.FunctionDef))
    assert (
        lookup_type(ti, return_node, return_node.value.name)
        == List[lookup_type(ti, functiondef_node, "x")]
    )



from typing import List

from astroid import nodes
from hypothesis import HealthCheck, given, settings

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


@given(cs.homogeneous_iterable)
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_list_comprehension_single_target_name_homogeneous_iterable(iterable):
    """Test Comprehension node visitor representing a comprehension expression with a single target and a
    name expression over a homogeneous list."""
    program = f"[num for num in {repr(iterable)}]"
    module, typeinferrer = cs._parse_text(program)
    listcomp_node = list(module.nodes_of_class(nodes.ListComp))[0]
    expected_type = List[listcomp_node.generators[0].iter.inf_type.getValue().__args__[0]]
    assert listcomp_node.inf_type.getValue() == expected_type


@given(cs.homogeneous_iterable)
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_list_comprehension_single_target_name_heterogeneous_iterable(iterable):
    """Test Comprehension node visitor representing a comprehension expression with a single target and a
    name expression over a heterogeneous list."""
    program = f"[num for num in {repr(iterable)}]"
    module, typeinferrer = cs._parse_text(program)
    listcomp_node = list(module.nodes_of_class(nodes.ListComp))[0]
    expected_type = List[listcomp_node.generators[0].iter.inf_type.getValue().__args__[0]]
    assert listcomp_node.inf_type.getValue() == expected_type


@given(cs.valid_identifier(min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_list_comprehension_single_target_name_string(iterable):
    """Test Comprehension node visitor representing a comprehension expression with a single target and a
    name expression over a string."""
    program = f"[num for num in {repr(iterable)}]"
    module, typeinferrer = cs._parse_text(program)
    listcomp_node = list(module.nodes_of_class(nodes.ListComp))[0]
    expected_type = List[listcomp_node.generators[0].iter.inf_type.getValue()]
    assert listcomp_node.inf_type.getValue() == expected_type



import pycodestyle
from pylint.checkers import BaseChecker
from pylint.interfaces import IRawChecker


class PycodestyleChecker(BaseChecker):
    __implements__ = IRawChecker

    name = "pep8_errors"
    msgs = {"E9989": ("Found pycodestyle (PEP8) style error at %s", "pep8-errors", "")}

    options = (
        (
            "pycodestyle-ignore",
            {
                "default": (),
                "type": "csv",
                "metavar": "<pycodestyle-ignore>",
                "help": "List of Pycodestyle errors to ignore",
            },
        ),
    )

    # this is important so that your checker is executed before others
    priority = -1

    def process_module(self, node):
        style_guide = pycodestyle.StyleGuide(
            paths=[node.stream().name], reporter=JSONReport, ignore=self.config.pycodestyle_ignore
        )
        report = style_guide.check_files()

        for line_num, msg in report.get_file_results():
            self.add_message("pep8-errors", line=line_num, args=msg)


class JSONReport(pycodestyle.StandardReport):
    def get_file_results(self):
        self._deferred_print.sort()
        return [
            (line_number, f"line {line_number}, column {offset}: {text}")
            for line_number, offset, _, text, _ in self._deferred_print
        ]


def register(linter):
    """required method to auto register this checker"""
    linter.register_checker(PycodestyleChecker(linter))



from typing import Any, Set

from astroid import nodes
from hypothesis import HealthCheck, assume, given, settings

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


@given(cs.simple_homogeneous_set_node())
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_homogeneous_set(node):
    """Test Set nodes representing a set of homogeneous values."""
    module, _ = cs._parse_text(node)
    set_node = list(module.nodes_of_class(nodes.Set))[0]
    if len(set_node.elts) == 0:
        assert set_node.inf_type.getValue() == Set[Any]
    else:
        try:
            cs._verify_type_setting(module, nodes.Set, Set[type(set_node.elts[0].value)])
        except AttributeError:
            cs._verify_type_setting(module, nodes.Set, Set[type(set_node.elts[0].operand.value)])


@given(cs.set_node(min_size=2))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_random_set(node):
    """Test Set nodes representing a set of heterogeneous values."""
    assume(not isinstance(list(node.elts)[0].value, type(list(node.elts)[1].value)))
    assume(not isinstance(list(node.elts)[1].value, type(list(node.elts)[0].value)))
    val_types = [type(val.value) for val in node.elts]
    if int in val_types:
        assume(bool not in val_types)
    if bool in val_types:
        assume(int not in val_types)
    module, _ = cs._parse_text(node)
    set_node = list(module.nodes_of_class(nodes.Set))[0]
    cs._verify_type_setting(module, nodes.Set, Set[Any])



import inspect

from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.checkers.utils import check_messages
from pylint.interfaces import IAstroidChecker


class ForbiddenImportChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = "forbidden_import"
    msgs = {
        "E9999": (
            "You may not import any modules - you imported %s on line %s.",
            "forbidden-import",
            "Used when you use import",
        )
    }
    options = (
        (
            "allowed-import-modules",
            {
                "default": (),
                "type": "csv",
                "metavar": "<modules>",
                "help": "Allowed modules to be imported.",
            },
        ),
        (
            "extra-imports",
            {
                "default": (),
                "type": "csv",
                "metavar": "<extra-modules>",
                "help": "Extra allowed modules to be imported.",
            },
        ),
    )

    # this is important so that your checker is executed before others
    priority = -1

    @check_messages("forbidden-import")
    def visit_import(self, node):
        """visit an Import node"""
        temp = [
            name
            for name in node.names
            if name[0] not in self.config.allowed_import_modules
            and name[0] not in self.config.extra_imports
        ]

        if temp != []:
            self.add_message(
                "forbidden-import",
                node=node,
                args=(", ".join(map(lambda x: x[0], temp)), node.lineno),
            )

    @check_messages("forbidden-import")
    def visit_importfrom(self, node):
        """visit an ImportFrom node"""
        if (
            node.modname not in self.config.allowed_import_modules
            and node.modname not in self.config.extra_imports
        ):
            self.add_message("forbidden-import", node=node, args=(node.modname, node.lineno))

    @check_messages("forbidden-import")
    def visit_call(self, node):
        if isinstance(node.func, nodes.Name):
            name = node.func.name
            # ignore the name if it's not a builtin (i.e. not defined in the
            # locals nor globals scope)
            if not (name in node.frame() or name in node.root()):
                if name == "__import__":
                    if (
                        node.args[0].value not in self.config.allowed_import_modules
                        and node.args[0].value not in self.config.extra_imports
                    ):
                        args = (node.args[0].value, node.lineno)
                        # add the message
                        self.add_message("forbidden-import", node=node, args=args)


def register(linter):
    """required method to auto register this checker"""
    linter.register_checker(ForbiddenImportChecker(linter))



from typing import TypeVar

from astroid import nodes
from pytest import skip

from python_ta.typecheck.base import TypeFailFunction

from .. import custom_hypothesis_support as cs


def test_overload_function():
    program = """
    def foo(x, y=None):
        return x + 5

    foo(5)
    foo(5, 6)
    """
    ast_mod, ti = cs._parse_text(program)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert call_node.inf_type.getValue() == int


def test_overload_function_2():
    program = """
    def foo(x, y=None, z=None):
        return x + 5
    foo(5)
    foo(5, 6)
    foo(5, 6, 7)
    """
    ast_mod, ti = cs._parse_text(program)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert call_node.inf_type.getValue() == int


def test_overload_function_with_gap():
    program = """
    def foo(x, y=None, z=None):
        return x + 5
    foo(5, None, 7)
    """
    ast_mod, ti = cs._parse_text(program)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert call_node.inf_type.getValue() == int


def test_too_few_args():
    program = """
       def foo(x, y):
           return x + 5
       foo(5)
       """
    ast_mod, ti = cs._parse_text(program)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert isinstance(call_node.inf_type, TypeFailFunction)


def test_too_few_args_2():
    program = """
       def foo(x, y, z):
           return x + 5
       foo(5, 6)
       """
    ast_mod, ti = cs._parse_text(program)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert isinstance(call_node.inf_type, TypeFailFunction)


def test_too_many_args():
    program = """
       def foo(x):
           return x + 5
       foo(5, 6)
       """
    ast_mod, ti = cs._parse_text(program)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert isinstance(call_node.inf_type, TypeFailFunction)


def test_too_many_args_2():
    program = """
       def foo(x, y):
           return x + 5
       foo(5, 6, 7)
       """
    ast_mod, ti = cs._parse_text(program)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert isinstance(call_node.inf_type, TypeFailFunction)


def test_too_few_args_with_overload():
    program = """
       def foo(x, y, z=None):
           return x + 5
       foo(5)
       """
    ast_mod, ti = cs._parse_text(program)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert isinstance(call_node.inf_type, TypeFailFunction)


def test_too_many_args_with_overload():
    program = """
       def foo(x, y=None):
           return x + 5
       foo(5, 6, 7)
       """
    ast_mod, ti = cs._parse_text(program)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert isinstance(call_node.inf_type, TypeFailFunction)


def test_overload_function_with_annotations():
    program = """
    def foo(x: int, y: int=None):
        return x + 5
    foo(5)
    foo(5, 6)
    """
    ast_mod, ti = cs._parse_text(program)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert call_node.inf_type.getValue() == int


def test_flagged_builtin_overload():
    program = """
    x = round(5.5)
    y = round(5.5, 1)
    """
    ast_mod, ti = cs._parse_text(program)
    for assgn_node in ast_mod.nodes_of_class(nodes.AssignName):
        if assgn_node.name == "x":
            x = ti.lookup_typevar(assgn_node, assgn_node.name)
            assert ti.type_constraints.resolve(x).getValue() == int
        if assgn_node.name == "y":
            y = ti.lookup_typevar(assgn_node, assgn_node.name)
            assert ti.type_constraints.resolve(y).getValue() == float


def test_builtin_defaults():
    program = """
    x = range(1)
    y = range(1, 10)
    z = range(1, 10, 3)
    """
    ast_mod, ti = cs._parse_text(program)
    for assgn_node in ast_mod.nodes_of_class(nodes.AssignName):
        x = ti.lookup_typevar(assgn_node, assgn_node.name)
        assert ti.type_constraints.resolve(x).getValue() == range


def test_unresolved_builtin():
    skip("Requires proper handling of builtins with multiple signatures")
    program = """
    def f(x, y):
        return x + y

    z = f('abc', 'def')
    """
    ast_mod, ti = cs._parse_text(program)

    for assgn_node in ast_mod.nodes_of_class(nodes.AssignName):
        if assgn_node.name == "z":
            z = ti.lookup_typevar(assgn_node, assgn_node.name)
            assert ti.type_constraints.resolve(z).getValue() == str


def test_unresolved_builtin2():
    skip("Requires proper handling of builtins with multiple signatures")
    program = """
    def f(x, y):
        return x + y

    def g(x, y):
        z = f(x, y)
        return z + 'abc'
    """
    ast_mod, ti = cs._parse_text(program)

    f_ret_node, g_ret_node = ast_mod.nodes_of_class(nodes.Return)
    assert isinstance(f_ret_node, TypeVar)
    assert g_ret_node.inf_type.getValue() == str



"""checker for a loop that can only ever run for one iteration.
"""
from typing import Union

from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.checkers.utils import check_messages
from pylint.interfaces import IAstroidChecker


class OneIterationChecker(BaseChecker):

    __implements__ = IAstroidChecker
    # name is the same as file name but without _checker part
    name = "one_iteration"
    # use dashes for connecting words in message symbol
    msgs = {
        "E9996": (
            "This loop will only ever run for one iteration",
            "one-iteration",
            "Reported when the loop body always ends the loop in its first iteration "
            '(e.g., by returning or using the "break" keyword).',
        )
    }

    # this is important so that your checker is executed before others
    priority = -1

    # pass in message symbol as a parameter of check_messages
    @check_messages("one-iteration")
    def visit_for(self, node):
        if self._check_one_iteration(node):
            self.add_message("one-iteration", node=node)

    @check_messages("one-iteration")
    def visit_while(self, node):
        if self._check_one_iteration(node):
            self.add_message("one-iteration", node=node)

    def _check_one_iteration(self, node: Union[nodes.For, nodes.While]) -> bool:
        """Return whether the given loop is guaranteed to stop after one iteration.

        More precisely, Returns False if there exists a direct predecessor
        block `p` to the start of the loop block `s` such that the
        first statement in `p` is a child node of <node> and that there exists a
        path from `s` to `p.

        Note: For `while` loops, 'start of the loop block' refers to the block with
        the test condition (or the first of the blocks that make up test condition).
        For `for` loops, it refers to the block with the assignment target.
        """
        start = node.target if isinstance(node, nodes.For) else node
        if not hasattr(start, "cfg_block"):
            return False

        preds = start.cfg_block.predecessors

        if preds == []:
            return False

        for pred in preds:
            stmt = pred.source.statements[0]
            if node.parent_of(stmt) and pred.source.reachable:
                if isinstance(node, nodes.For) and stmt is node.iter:
                    continue
                return False
        return True


def register(linter):
    linter.register_checker(OneIterationChecker(linter))



"""Checker for redundant assignment statements in the program.

An assignment statement is redundant if it satisfies the following two properties:

    1. Every path from a definition of a variable `v` to its first usage
    goes through at least one re-definition of `v`.

    2. Removing the statement from the program does not in any way change
    the behavior of the program.
"""
from typing import List, Set, Union

from astroid import nodes
from pylint.checkers import BaseChecker, utils
from pylint.checkers.utils import check_messages
from pylint.interfaces import IAstroidChecker

from python_ta.cfg.graph import CFGBlock, ControlFlowGraph


class RedundantAssignmentChecker(BaseChecker):

    __implements__ = IAstroidChecker
    # name is the same as file name but without _checker part
    name = "redundant_assignment"
    # use dashes for connecting words in message symbol
    msgs = {
        "E9959": (
            "This assignment statement is redundant;" " You can remove it from the program.",
            "redundant-assignment",
            "This assignment statement is redundant;" " You can remove it from the program.",
        )
    }

    # this is important so that your checker is executed before others
    priority = -1

    def __init__(self, linter=None):
        super().__init__(linter=linter)
        self._redundant_assignment: Set[nodes.Assign] = set()

    @check_messages("redundant-assignment")
    def visit_assign(self, node: nodes.Assign):
        if node in self._redundant_assignment:
            self.add_message("redundant-assignment", node=node)

    @check_messages("redundant-assignment")
    def visit_augassign(self, node: nodes.AugAssign):
        if node in self._redundant_assignment:
            self.add_message("redundant-assignment", node=node)

    def visit_module(self, node: nodes.Module):
        self._analyze(node)

    def visit_functiondef(self, node: nodes.FunctionDef):
        self._analyze(node)

    def _analyze(self, node: Union[nodes.Module, nodes.FunctionDef]) -> None:
        """Runs the backward data flow algorithm on a `Module` or `Function` CFG, which in turn
        appends `Assign` nodes to `self.redundant_assignment` if it is redundant.

        Data flow algorithms retrieved from:
        https://www.seas.harvard.edu/courses/cs252/2011sp/slides/Lec02-Dataflow.pdf#page=31
        """
        # Stores all the variable names that will be re-defined before any usage at a
        # particular program point.
        out_facts = {}
        cfg = ControlFlowGraph()
        cfg.start = node.cfg_block
        worklist = list(cfg.get_blocks_postorder())
        worklist.reverse()

        all_assigns = self._get_assigns(node)
        for block in worklist:
            out_facts[block] = all_assigns.copy()

        while len(worklist) != 0:
            b = worklist.pop()
            outs = [out_facts[p.target] for p in b.successors if p.target in out_facts]
            if outs == []:
                in_facts = set()
            else:
                in_facts = set.intersection(*outs)
            temp = self._transfer(b, in_facts)
            if temp != out_facts[b]:
                out_facts[b] = temp
                worklist.extend([pred.source for pred in b.predecessors if pred.source.reachable])

    def _transfer(self, block: CFGBlock, out_facts: Set[str]) -> Set[str]:
        gen = out_facts.copy()
        kill = set()
        for statement in reversed(block.statements):
            if isinstance(statement, nodes.FunctionDef):
                # `nodes_of_class` below doesnt block looking for required nodes
                # in function definitions, hence this case.
                continue
            for node in statement.nodes_of_class(
                (
                    nodes.AssignName,
                    nodes.DelName,
                    nodes.Name,
                    nodes.Nonlocal,
                    nodes.Global,
                ),
                nodes.FunctionDef,
            ):
                if isinstance(node, nodes.AssignName):
                    if node.name in gen.difference(kill):
                        self._redundant_assignment.add(node.parent)
                    elif node.parent in self._redundant_assignment:
                        self._redundant_assignment.remove(node.parent)

                    # When node.parent is an AugAssign, the name counts as a use of the variable,
                    # and so is added to kill.
                    if isinstance(node.parent, nodes.AugAssign):
                        kill.add(node.name)
                    else:
                        kill.discard(node.name)
                    gen.add(node.name)
                elif isinstance(node, (nodes.Nonlocal, nodes.Global)):
                    kill.difference_update(set(node.names))
                else:
                    kill.add(node.name)

        return gen.difference(kill)

    def _get_assigns(self, node: Union[nodes.FunctionDef, nodes.Module]) -> Set[str]:
        """Returns a set of all local and parameter variables that could be
        defined in the program (either a function or module).

        IF a variable 'v' is defined in a function and there is no global/nonlocal
        statement applied to 'v' THEN 'v' is a local variable.

        Note that `local variable` in the context of a module level analysis,
        refers to global variables.
        """
        assigns = set()
        kills = set()
        for name, assign_nodes in node.locals.items():
            if any(isinstance(elem, nodes.AssignName) for elem in assign_nodes):
                assigns.add(name)
        return assigns.difference(kills)


def register(linter):
    linter.register_checker(RedundantAssignmentChecker(linter))



from astroid import nodes
from hypothesis import settings
from pytest import skip

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


def test_incompatible_binop_call():
    """User tries to call a builtin binary operation on arguments of the wrong type."""
    skip("SKIP FOR NOW.")
    program = f'5 + "string"\n'
    module, _ = cs._parse_text(program)
    binop_node = next(module.nodes_of_class(nodes.BinOp))
    expected_msg = (
        "You cannot add an int, 5, and a str, 'string'. "
        "Perhaps you wanted to cast the integer into a string or vice versa?"
    )
    assert binop_node.inf_type.getValue() == expected_msg


def test_incompatible_unaryop_call():
    """User tries to call a builtin unary operation on an argument of the wrong type."""
    skip("SKIP FOR NOW.")
    program = f'~["D"]'
    module, _ = cs._parse_text(program)
    unaryop_node = next(module.nodes_of_class(nodes.UnaryOp))
    expected_msg = "You cannot take the bitwise inverse of a list of str, ['D']."
    assert unaryop_node.inf_type.getValue() == expected_msg


def test_incompatible_subscript_list():
    """User tries to access an element of a list using the wrong type of index."""
    skip("SKIP FOR NOW.")
    program = f'[1,2,3]["one"]'
    module, _ = cs._parse_text(program)
    subscript_node = next(module.nodes_of_class(nodes.Subscript))
    expected_msg = "You can only access elements of a list using an int. You used a str, 'one'."
    assert subscript_node.inf_type.getValue() == expected_msg


def test_incompatible_subscript_tuple():
    """User tries to access an element of a tuple using the wrong type of index."""
    skip("SKIP FOR NOW.")
    program = f'(1,2,3)["one"]'
    module, _ = cs._parse_text(program)
    subscript_node = next(module.nodes_of_class(nodes.Subscript))
    expected_msg = "You can only access elements of a tuple using an int. You used a str, 'one'."
    assert subscript_node.inf_type.getValue() == expected_msg


def test_incompatible_subscript_dictionary():
    """User tries to access an element of a dictionary using the wrong type of key."""
    skip("SKIP FOR NOW.")
    program = """{ "1" : 1, "2" : 2, "3" : 3 }[1]"""
    module, _ = cs._parse_text(program)
    subscript_node = next(module.nodes_of_class(nodes.Subscript))
    expected_msg = "You tried to access an element of this dictionary using an int, 1, but the keys are of type str."
    assert subscript_node.inf_type.getValue() == expected_msg



import sys
import typing
from itertools import product
from typing import *
from typing import IO  # Needed for type_store
from typing import Callable, ForwardRef, _GenericAlias, _type_check

from astroid import nodes

from python_ta.typecheck.errors import error_message
from python_ta.utils import _get_name, _gorg

from ..util.monad import Failable, failable_collect


class _TNode:
    """A node in the TypeConstraints disjoint set data structure."""

    type: type
    parent: Optional["_TNode"]
    parent_path: Optional[List[nodes.NodeNG]]
    adj_list: List[Tuple["_TNode", nodes.NodeNG]]
    ast_node: Optional[nodes.NodeNG]

    def __init__(self, node_type: type, ast_node: Optional[nodes.NodeNG] = None) -> None:
        self.type = node_type
        self.parent = None
        self.parent_path = None
        self.adj_list = []
        self.ast_node = ast_node

    def __eq__(self, other: "_TNode") -> bool:
        if str(self.type) == str(other.type):
            return True
        else:
            return False

    def __str__(self) -> str:
        if self.parent and self.ast_node:
            return f"TNode {self.ast_node.as_string()}: {self.type}, resolved to {self.parent.type}"
        elif self.ast_node:
            return f"TNode {self.ast_node.as_string()}: {self.type}"
        else:
            return f"TNode: {self.type}"

    def find_path_to_parent(self) -> List[nodes.NodeNG]:
        """Return list of astroid nodes relating _TNode to parent _TNode."""
        final_path = []
        cur_node = self
        while cur_node.parent_path:
            final_path.append(cur_node.parent_path[1])
            cur_node = cur_node.parent_path[0]
        return final_path

    def find_annotation(self) -> Optional[nodes.AnnAssign]:
        """Find annotation node in list of astroid nodes relating _TNode to parent _TNode, if one exists."""
        path = self.find_path_to_parent()
        for p in path:
            if isinstance(p, nodes.AnnAssign):
                return p


class TypeResult(Failable):
    """Represents the result of a type check operation that either succeeded or
    failed.
    """

    def __init__(self, value) -> None:
        super().__init__(value)


class TypeInfo(TypeResult):
    """Represents the result of a successful type check operation
    Contains information about the inferred type of a node
    """

    def __init__(self, type_: type) -> None:
        super().__init__(type_)

    def __repr__(self) -> str:
        return f"TypeInfo({repr(self.value)})"

    def __str__(self) -> str:
        return f"TypeInfo: {self.value}"


class NoType(TypeResult):
    """Class representing no inferred type"""

    def __init__(self) -> None:
        super().__init__(None)


class TypeFail(TypeResult):
    """Represents the result of a failed type check operation.
    Contains error message.
    """

    def __init__(self, msg: Optional[str] = None) -> None:
        self.msg = msg
        super().__init__(self.msg)

    def __str__(self) -> str:
        return f"TypeFail: {self.msg}"

    def bind(self, _) -> "TypeFail":
        return self


class TypeFailUnify(TypeFail):
    """
    TypeFailUnify occurs when two types fail to unify.

    :param tnodes: List of _TNodes that failed to unify. Usually contains two
    :param src_node: astroid node where failure occurs
    """

    def __init__(self, *tnodes: _TNode, src_node: nodes.NodeNG = None) -> None:
        self.tnodes = tnodes
        self.src_node = src_node
        super().__init__(str(self))

    def __str__(self) -> str:
        string = "TypeFail: Unable to Unify "
        string += (
            f"{self.tnodes[0].ast_node.as_string()}"
            if self.tnodes[0].ast_node
            else f"{self.tnodes[0].type}"
        )
        string += " <-> "
        string += (
            f"{self.tnodes[1].ast_node.as_string()}"
            if self.tnodes[1].ast_node
            else f"{self.tnodes[1].type}"
        )
        if self.src_node:
            string += f" at {self.src_node.as_string()}"
        return string


class TypeFailLookup(TypeFail):
    """
    TypeFailLookup occurs when an attribute variable or method is called, and either the attribute or
    class is invalid.

    :param class_tnode: _TNode of looked up class
    :param attr_node: astroid node representing looked up attribute
    :param src_node: astroid node where invalid lookup occurs
    """

    def __init__(
        self, class_tnode: _TNode, attr_node: nodes.NodeNG, src_node: nodes.NodeNG
    ) -> None:
        self.class_tnode = class_tnode
        self.attr_node = attr_node
        self.src_node = src_node
        super().__init__(str(self))

    def __str__(self) -> str:
        return f"TypeFail: Invalid attribute lookup {self.src_node.as_string()}"


class TypeFailAnnotationUnify(TypeFail):
    """
    TypeFailAnnotationUnify occurs when a contradiction occurs during the unification of the inferred type
    and the annotated type.

    :param tnode: _TNode of expected type
    :param src_node: astroid node where error occurs
    :param ann_node: astroid node where annotation is set
    """

    def __init__(
        self, tnode: _TNode, src_node: nodes.NodeNG = None, ann_node: nodes.NodeNG = None
    ) -> None:
        self.tnode = tnode
        self.src_node = src_node
        self.ann_node = ann_node
        super().__init__(str(self))

    def __str__(self) -> str:
        string = f"TypeFail: Annotation error in {self.src_node.as_string()}. {self.tnode.ast_node.as_string()} is annotated as "
        string += (
            f"{_get_name(self.tnode.parent.type)}"
            if self.tnode.parent
            else f"{_get_name(self.tnode.type)}"
        )
        string += f" at {self.ann_node.as_string()}"
        return string


class TypeFailAnnotationInvalid(TypeFail):
    """
    TypeFailAnnotationInvalid occurs when a variable is annotated as something other than a type

    :param src_node: astroid node where annotation is set
    """

    def __init__(self, src_node: nodes.NodeNG) -> None:
        self.src_node = src_node
        super().__init__(str(self))

    def __str__(self) -> str:
        return f"TypeFail: Annotation must be a type"


class TypeFailFunction(TypeFail):
    """
    TypeFailFunction occurs when a function is called with different arguments than expected.

    :param func_types: Tuple containing one or more acceptable type signatures
    :param funcdef_node: FunctionDef astroid node where function is defined
    :param src_node: Astroid node where invalid function call occurs
    :param arg_indices: List of argument index numbers,
    """

    def __init__(
        self,
        func_types: Tuple[Callable],
        funcdef_node: nodes.FunctionDef,
        src_node: nodes.NodeNG,
        arg_indices: List[int] = None,
    ) -> None:
        self.func_types = func_types
        self.funcdef_node = funcdef_node
        self.src_node = src_node
        self.arg_indices = arg_indices
        super().__init__(str(self))

    def __str__(self):
        return error_message(self)


class TypeFailReturn(TypeFail):
    """
    TypeFailReturn occurs when a nodes.Return node is encountered outside of a function definition.

    :param src_node: Invalid nodes.Return node
    """

    def __init__(self, src_node: nodes.Return) -> None:
        self.src_node = src_node
        super().__init__(str(self))

    def __str__(self) -> str:
        return f"TypeFail: Return statement not in valid function"


class TypeFailStarred(TypeFail):
    """
    TypeFailStarred occurs when there are multiple starred variables in the target of an assignment.

    :param src_node: Invalid nodes.Assign node
    """

    def __init__(self, src_node: nodes.Assign) -> None:
        self.src_node = src_node
        super().__init__(str(self))

    def __str__(self) -> str:
        return f"TypeFail: Multiple starred variables not valid"


def accept_failable(f: Callable) -> Callable:
    """Decorator to allow function fo to optionally acceptance instances of Failable as arguments."""

    def _f(*args, **kwargs):
        """Extract value from Failable arguments, pass values to function f.
        Return TypeFail instead if found.
        """
        new_args = []
        new_kwargs = {}
        for a in args:
            if isinstance(a, Failable):
                if isinstance(a, TypeFail):
                    return a
                a >> new_args.append
            else:
                new_args.append(a)
        for kw in kwargs:
            if isinstance(kwargs[kw], Failable):
                if isinstance(kwargs[kw], Failable):
                    return kwargs[kw]
                new_kwargs += kwargs[kw] >> (lambda t: dict(kw=t))
            else:
                new_kwargs[kw] = kwargs[kw]
        return f(*new_args, **new_kwargs)

    return _f


@accept_failable
def _wrap_generic_meta(t: _GenericAlias, args: List[type]) -> TypeResult:
    if t.__origin__ is tuple:
        tuple_args = tuple(args)
        # Handle the special case when t1 or t2 are empty tuples; TODO: investigate this
        if tuple_args == ((),):
            tuple_args = ()
        return TypeInfo(Tuple[tuple_args])
    elif is_callable(t):
        c = Callable.copy_with(tuple(args))
        c.__polymorphic_tvars__ = getattr(t, "__polymorphic_tvars__", frozenset())
        return TypeInfo(c)
    else:
        return TypeInfo(t.copy_with(tuple(args)))


@accept_failable
def wrap_container(container_type: _GenericAlias, *args: type) -> TypeResult:
    """Return instance of type container_type with type variable arguments args, wrapped in instance of TypeInfo."""
    return TypeInfo(container_type.copy_with(args))


Num = TypeVar("number", int, float)
a = TypeVar("a")
MulNum = TypeVar("mul_n", int, float, str, List[a])
tup1 = TypeVar("tup1")
tup2 = TypeVar("tup2")


class TuplePlus(TypeVar, _root=True):
    def eval_type(self, type_constraints: "TypeConstraints") -> TypeResult:
        t1, t2 = self.__constraints__
        t1 = type_constraints.resolve(t1).__params__
        t2 = type_constraints.resolve(t2).__params__
        return wrap_container(Tuple, t1, t2)


_TYPESHED_TVARS = {
    "_T": TypeVar("_T"),
    "_T_co": TypeVar("_T_co", covariant=True),
    "_KT": TypeVar("_KT"),
    "_VT": TypeVar("_VT"),
    "_S": TypeVar("_S"),
    "_T1": TypeVar("_T1"),
    "_T2": TypeVar("_T2"),
    "_T3": TypeVar("_T3"),
    "_T4": TypeVar("_T4"),
    "_T5": TypeVar("_T5"),
    "_TT": TypeVar("_TT", bound="type"),
    "function": Callable[[Any], Any],
}


_BUILTIN_TO_TYPING = {
    "list": "List",
    "dict": "Dict",
    "tuple": "Tuple",
    "set": "Set",
    "frozenset": "FrozenSet",
    "function": "Callable",
    "Iterator": "Iterator",
}


def _get_poly_vars(t: type) -> Set[str]:
    """Return a set consisting of the names of all TypeVars within t"""
    if isinstance(t, TypeVar) and t.__name__ in _TYPESHED_TVARS:
        return set([t.__name__])
    elif isinstance(t, _GenericAlias) and t.__args__:
        pvars = set()
        for arg in t.__args__:
            pvars.update(_get_poly_vars(arg))
        return pvars
    return set()


def create_Callable(
    args: Iterable[type], rtype: type, class_poly_vars: Set[type] = None
) -> Callable:
    """Initialize and return Callable with given parameters, return types, and polymorphic type variables."""
    poly_vars = set(class_poly_vars) if class_poly_vars else set()
    c = Callable.copy_with(tuple([*args, rtype]))
    poly_vars.update(_get_poly_vars(c))
    c.__polymorphic_tvars__ = frozenset(poly_vars)
    return c


@accept_failable
def create_Callable_TypeResult(
    args: Iterable[type], rtype: type, poly_vars: Set[type] = None
) -> TypeResult:
    """Return Callable wrapped in a TypeInfo instance."""
    return TypeInfo(create_Callable(args, rtype, poly_vars))


TYPE_SIGNATURES = {
    int: {
        "__add__": create_Callable([int, Num], Num, {Num}),
        "__sub__": create_Callable([int, Num], Num, {Num}),
        "__mul__": create_Callable([int, MulNum], MulNum, {MulNum}),
        "__idiv__": create_Callable([int, Num], Num, {Num}),
        "__mod__": create_Callable([int, Num], Num, {Num}),
        "__pow__": create_Callable([int, Num], Num, {Num}),
        "__div__": create_Callable([int, Num], float, {Num}),
    },
    float: {
        "__add__": create_Callable([float, Num], float, {Num}),
        "__sub__": create_Callable([float, Num], float, {Num}),
        "__mul__": create_Callable([float, Num], float, {Num}),
        "__idiv__": create_Callable([float, Num], float, {Num}),
        "__mod__": create_Callable([float, Num], float, {Num}),
        "__pow__": create_Callable([float, Num], float, {Num}),
        "__div__": create_Callable([float, Num], float, {Num}),
    },
    str: {"__add__": Callable[[str, str], str], "__mul__": Callable[[str, int], str]},
    List: {
        "__add__": create_Callable([List[a], List[a]], List[a], {a}),
        "__mul__": create_Callable([List[a], int], List[a], {a}),
        "__getitem__": create_Callable([List[a], int], a, {a}),
    },
    Tuple: {"__add__": create_Callable([tup1, tup2], TuplePlus("tup+", tup1, tup2), {tup1, tup2})},
}


def op_to_dunder_binary(op: str) -> str:
    """Return the dunder method name corresponding to binary op."""
    if op == "+":
        return "__add__"
    elif op == "-":
        return "__sub__"
    elif op == "*":
        return "__mul__"
    elif op == "//":
        return "__idiv__"
    elif op == "%":
        return "__mod__"
    elif op == "/":
        return "__div__"
    elif op == "**":
        return "__pow__"
    elif op == "&":
        return "__and__"
    elif op == "^":
        return "__xor__"
    elif op == "|":
        return "__or__"
    elif op == "==":
        return "__eq__"
    elif op == "!=":
        return "__ne__"
    elif op == "<":
        return "__lt__"
    elif op == "<=":
        return "__le__"
    elif op == ">":
        return "__gt__"
    elif op == ">=":
        return "__ge__"
    # TODO: 'is' and 'in'
    else:
        return op


def op_to_dunder_unary(op: str) -> str:
    """Return the dunder method name corresponding to unary op."""
    if op == "-":
        return "__neg__"
    elif op == "+":
        return "__pos__"
    elif op == "~":
        return "__invert__"
    else:
        return op


class TypeConstraints:
    """Represents all the type constraints in the program.

    This is mainly comprised of a disjoint set data structure, in which each disjoint set
    represents a set of equivalences of type variables and concrete types. The nodes
    in the disjoint set are implemented by the private class _TNode above.
    """

    # The number of type variables stored in the data structure. Used to generate fresh type variables.
    _count: int
    # List of _TNodes
    _nodes: List[_TNode]
    # A mapping of types to nodes
    type_to_tnode: Dict[str, _TNode]

    def __init__(self) -> None:
        self.type_store = None
        self.reset()

    def __deepcopy__(self, memodict: Dict = {}) -> "TypeConstraints":
        tc = TypeConstraints()
        tc._count = self._count
        tc._nodes = []
        tc.type_to_tnode = {}
        tc.type_store = self.type_store
        # copy nodes without copying edges
        for node in self._nodes:
            node_cpy = _TNode(node.type, node.ast_node)
            tc._nodes.append(node_cpy)
            tc.type_to_tnode[str(node.type)] = node_cpy
        # fill in edges
        for node in self._nodes:
            for adj_node, ctx in node.adj_list:
                tc.type_to_tnode[str(node.type)].adj_list.append(
                    (tc.type_to_tnode[str(adj_node.type)], ctx)
                )
            if node.parent:
                tc.type_to_tnode[str(node.type)].parent = tc.type_to_tnode[str(node.parent.type)]
        return tc

    def reset(self) -> None:
        """Reset the type constraints kept track of in the program."""
        self._count = 0
        self._nodes = []
        self.type_to_tnode = {}

    ###########################################################################
    # Creating new nodes ("make set")
    ###########################################################################
    # TODO: Rename to better distinguish between _TNodes and AST Nodes
    def fresh_tvar(self, node: Optional[nodes.NodeNG] = None) -> TypeVar:
        """Create and return a fresh type variable, associated with the given node."""
        tvar = TypeVar(f"_TV{self._count}")
        self._count += 1
        self._make_set(tvar, ast_node=node)
        return tvar

    def _make_set(self, t: type, ast_node: Optional[nodes.NodeNG] = None) -> _TNode:
        """Create new set with a single _TNode."""
        node = _TNode(t, ast_node)
        self._nodes.append(node)
        self.type_to_tnode[str(t)] = node
        if not isinstance(t, TypeVar):
            node.parent = node
        return node

    def get_tnode(self, t: type) -> _TNode:
        """Return the _TNode that represents the given type t, or create a new one."""
        try:
            node = self.type_to_tnode[str(t)]
        except KeyError:
            node = self._make_set(t, None)
        return node

    ###########################################################################
    # Type lookup ("find")
    ###########################################################################
    @accept_failable
    def resolve(self, t: type) -> TypeResult:
        """Return the concrete type or set representative associated with the given type."""
        if isinstance(t, _GenericAlias):
            if t.__args__ is not None:
                res_args = [self.resolve(arg) for arg in t.__args__]
                return _wrap_generic_meta(t, failable_collect(res_args))
            else:
                return TypeInfo(t)
        elif isinstance(t, TypeVar):
            try:
                repr = self.find_repr(self.type_to_tnode[str(t)])
                if repr and repr.type is not t:
                    if any(elt is t for elt in getattr(repr.type, "__args__", [])):
                        return TypeInfo(t)
                    else:
                        return self.resolve(repr.type)
            except KeyError:
                return TypeInfo(t)
        return TypeInfo(t)

    def is_concrete(self, t: type) -> bool:
        if isinstance(t, _GenericAlias):
            return all([self.is_concrete(arg) for arg in t.__args__])
        else:
            return not isinstance(t, TypeVar)

    def find_repr(self, tn: _TNode) -> Optional[_TNode]:
        """Search, using BFS starting from this _TNode, to find a _TNode that has a parent,
        or a unique set representative if no parent is found."""
        return self.find_parent(tn, True)

    def find_parent(self, tn: _TNode, find_repr: bool = False) -> Optional[_TNode]:
        """Search, using BFS starting from this _TNode, to find a _TNode that has a parent."""
        if tn.parent is not None:
            return tn.parent

        goal_tnode = self.find_node(tn, (lambda t: t.parent is not None), find_repr)

        if goal_tnode and goal_tnode.parent:
            cur_node = goal_tnode
            while cur_node:
                next_node = None
                for e in cur_node.adj_list:
                    if not e[0].parent_path and not e[0].parent:
                        e[0].parent = goal_tnode
                        e[0].parent_path = (cur_node, e[1])
                        next_node = e[0]
                cur_node = next_node

        return goal_tnode

    def find_function_def(self, tn: _TNode) -> Optional[nodes.FunctionDef]:
        """Search, using BFS starting from this _TNode, to find a _TNode with a
        FunctionDef node as its ast_node attribute."""
        func_tnode = self.find_node(
            tn, (lambda t: isinstance(t.ast_node, nodes.FunctionDef)), False
        )
        if func_tnode:
            return func_tnode.ast_node

    def find_node(
        self, tn: _TNode, cond: Callable[[Any], bool], find_repr: bool = False
    ) -> Optional[_TNode]:
        """Search, using BFS starting from this _TNode, to find a _TNode that satisfied passed in condition function."""
        visited = []
        node_list = [tn]
        goal_tnode = None
        while node_list:
            cur_node = node_list[0]
            for e in cur_node.adj_list:
                if e[0] not in visited and e[0] not in node_list:
                    if cond(e[0]):
                        goal_tnode = e[0]
                        break
                    node_list.append(e[0])
            visited.append(node_list[0])
            node_list.remove(node_list[0])

        # Return a set representative, even if it isn't a concrete type
        if find_repr and not goal_tnode and len(visited) > 1:
            visited_types = list(tnode.type for tnode in visited)
            visited_types.sort(key=(lambda t: t.__name__))
            goal_tnode = self.get_tnode(visited_types[-1])

        return goal_tnode

    def create_edges(self, tn1: _TNode, tn2: _TNode, ast_node: nodes.NodeNG):
        if tn1 != tn2:
            edge_exists = False
            for e in tn1.adj_list:
                if e[0] == tn2:
                    edge_exists = True
            if not edge_exists:
                tn1.adj_list.append((tn2, ast_node))
                tn2.adj_list.append((tn1, ast_node))

    ###########################################################################
    # Type unification ("union")
    ###########################################################################
    @accept_failable
    def unify(self, t1: type, t2: type, ast_node: Optional[nodes.NodeNG] = None) -> TypeResult:
        """Attempt to unify two types.

        :param t1: The first of the two types to be unified.
        :param t2: The second of the two types to be unified.
        :param ast_node: The astroid node responsible for the unification of t1 & t2.
        :returns: A TypeResult object (TypeFail or TypeInfo) containing information
            about the success / failure of the type unification.
        """

        # Get associated TNodes
        tnode1 = self.get_tnode(t1)
        tnode2 = self.get_tnode(t2)

        # Attempt to resolve to a TNode with concrete type
        conc_tnode1 = self.find_parent(tnode1)
        conc_tnode2 = self.find_parent(tnode2)

        # Both types can be resolved
        if conc_tnode1 is not None and conc_tnode2 is not None:
            ct1 = conc_tnode1.type
            ct2 = conc_tnode2.type
            if ct1 == ct2:
                tnode1.parent = conc_tnode1
                tnode2.parent = conc_tnode1
                self.create_edges(tnode1, tnode2, ast_node)
                return TypeInfo(ct1)
            elif (
                getattr(ct1, "__origin__", None) is Union
                or getattr(ct2, "__origin__", None) is Union
            ):
                ct1_types = ct1.__args__ if getattr(ct1, "__origin__", None) is Union else [ct1]
                ct2_types = ct2.__args__ if getattr(ct2, "__origin__", None) is Union else [ct2]
                for u1, u2 in product(ct1_types, ct2_types):
                    if self.can_unify(u1, u2):
                        return self.unify(u1, u2, ast_node)
                return TypeFailUnify(tnode1, tnode2, src_node=ast_node)
            elif isinstance(ct1, _GenericAlias) and isinstance(ct2, _GenericAlias):
                return self._unify_generic(tnode1, tnode2, ast_node)

            elif ct1 == Any or ct2 == Any:
                return TypeInfo(ct1) if ct1 != Any else TypeInfo(ct2)
            # Handle inheritance
            elif (
                self.type_store
                and ct1 is not None
                and ct2 is not None
                and self.type_store.is_descendant(ct1, ct2)
            ):
                return TypeInfo(ct1)
            else:
                for tn in [tnode1, tnode2]:
                    ann_t = tn.find_annotation()
                    if ann_t is not None:
                        return TypeFailAnnotationUnify(tn, ast_node, ann_t)
                return TypeFailUnify(tnode1, tnode2, src_node=ast_node)

        # One type can be resolved
        elif conc_tnode1 is not None:
            tnode2.parent = conc_tnode1
            tnode2.parent_path = (tnode1, ast_node)
            self.create_edges(tnode1, tnode2, ast_node)
            return TypeInfo(conc_tnode1.type)
        elif conc_tnode2 is not None:
            return self.unify(t2, t1, ast_node)

        # Both types are type variables
        elif t1 == t2:
            return TypeInfo(t1)
        else:
            self.create_edges(tnode1, tnode2, ast_node)
            return NoType()

    def _unify_generic(
        self, tnode1: _TNode, tnode2: _TNode, ast_node: Optional[nodes.NodeNG] = None
    ) -> TypeResult:
        """Unify two generic types (e.g., List, Tuple, Dict, Callable)."""

        conc_tnode1 = self.find_parent(tnode1)
        conc_tnode2 = self.find_parent(tnode2)

        g1, g2 = _gorg(conc_tnode1.type), _gorg(conc_tnode2.type)
        if not self.type_store or not self.type_store.is_descendant(
            conc_tnode1.type, conc_tnode2.type
        ):
            if (
                g1 is not g2
                or conc_tnode1.type.__args__ is None
                or conc_tnode2.type.__args__ is None
            ):
                # TODO: need to store more info here and in the case below
                return TypeFailUnify(conc_tnode1, conc_tnode2, src_node=ast_node)
            if len(conc_tnode1.type.__args__) != len(conc_tnode2.type.__args__):
                return TypeFailUnify(conc_tnode1, conc_tnode2, src_node=ast_node)

        arg_inf_types = []
        for a1, a2 in zip(conc_tnode1.type.__args__, conc_tnode2.type.__args__):
            if a1 == () or a2 == ():
                result = TypeInfo(())
            else:
                result = self.unify(a1, a2, ast_node)

            if isinstance(result, TypeFail):
                arg_inf_types = [TypeFailUnify(tnode1, tnode2, src_node=ast_node)]
                break
            else:
                arg_inf_types.append(result)

        unified_args = failable_collect(arg_inf_types)

        result = _wrap_generic_meta(conc_tnode1.type, unified_args)
        if not isinstance(result, TypeFail):
            self.create_edges(tnode1, tnode2, ast_node)
        return result

    ###########################################################################
    # Handling generic polymorphism
    ###########################################################################
    def can_unify(self, t1: type, t2: type) -> bool:
        """Check if the two types can unify without modifying current TypeConstraints."""
        tc = self.__deepcopy__()
        return not isinstance(tc.unify(t1, t2, None), TypeFail)

    @accept_failable
    def unify_call(
        self, func_var: type, *arg_types: type, node: Optional[nodes.NodeNG] = None
    ) -> TypeResult:
        """Unify a function call with the given function type and argument types.

        Return a result type.
        """
        if is_callable(func_var):
            func_type = func_var
        else:
            func_var_tnode = self.get_tnode(func_var)
            parent_tnode = self.find_parent(func_var_tnode)
            func_type = parent_tnode.type

        # Check for Callable[..., T]
        if is_callable(func_type) and func_type.__args__[0] is ...:
            return TypeInfo(func_type.__args__[-1])

        # Check that the number of parameters matches the number of arguments.
        if func_type.__origin__ is Union:
            new_func_type = None
            for c in func_type.__args__:
                if len(c.__args__) - 1 == len(arg_types):
                    new_func_type = c
                    break
            if new_func_type is None:
                func_var_tnode = self.get_tnode(func_var)
                funcdef_node = self.find_function_def(func_var_tnode)
                return TypeFailFunction(tuple(func_type.__args__), funcdef_node, node)
            else:
                func_type = new_func_type
        elif len(func_type.__args__) - 1 != len(arg_types):
            func_var_tnode = self.get_tnode(func_var)
            funcdef_node = self.find_function_def(func_var_tnode)
            return TypeFailFunction((func_type,), funcdef_node, node)
        new_func_type = self.fresh_callable(func_type, node)
        func_params = getattr(new_func_type, "__args__", [None])[:-1]
        arg_types = list(arg_types)
        for param, i in zip(func_params, range(len(arg_types))):
            # Automatic conversion of iterable types to Iterable[T]
            if isinstance(param, _GenericAlias) and param.__origin__ is Iterable.__origin__:
                arg = self.resolve(arg_types[i])
                func_type = self.type_store.lookup_method("__iter__", arg.getValue(), node=node)

                if isinstance(func_type, TypeFail):
                    func_var_tnode = self.get_tnode(func_var)
                    funcdef_node = self.find_function_def(func_var_tnode)
                    return TypeFailFunction((func_type,), funcdef_node, node)

                iterator_type = self.unify_call(func_type, arg_types[i], node=node)
                if isinstance(iterator_type, TypeFail):
                    func_var_tnode = self.get_tnode(func_var)
                    funcdef_node = self.find_function_def(func_var_tnode)
                    return TypeFailFunction((func_type,), funcdef_node, node)

                arg_types[i] = Iterable[iterator_type.getValue().__args__[0]]

        results = []
        for i in range(len(arg_types)):
            result = self.unify(arg_types[i], new_func_type.__args__[i], node)
            if isinstance(result, TypeFail):
                func_var_tnode = self.get_tnode(func_var)
                funcdef_node = self.find_function_def(func_var_tnode)
                param_annotations = funcdef_node.args.annotations if funcdef_node else None
                if param_annotations and param_annotations[i] is not None:
                    tvar = funcdef_node.type_environment.lookup_in_env(
                        funcdef_node.args.args[i].name
                    )
                    tnode = self.get_tnode(tvar)
                    return TypeFailAnnotationUnify(tnode, node, funcdef_node)
                else:
                    results.append(i)
        if results:
            func_var_tnode = self.get_tnode(func_var)
            funcdef_node = self.find_function_def(func_var_tnode)
            return TypeFailFunction((new_func_type,), funcdef_node, node, results)
        return self._type_eval(new_func_type.__args__[-1])

    def _type_eval(self, t: type) -> TypeResult:
        """Evaluate a type. Used for tuples."""
        if isinstance(t, TuplePlus):
            return t.eval_type(self)
        if isinstance(t, TypeVar):
            return self.resolve(t)
        if isinstance(t, _GenericAlias) and t.__args__ is not None:
            inf_args = (self._type_eval(argument) for argument in t.__args__)
            return wrap_container(t, *inf_args)
        else:
            return TypeInfo(t)

    def fresh_callable(self, func_type: type, node: Optional[nodes.NodeNG]) -> type:
        """Given a callable, substitute all polymorphic variables with fresh ones"""
        new_tvars = {
            tvar: self.fresh_tvar(node) for tvar in getattr(func_type, "__polymorphic_tvars__", [])
        }
        return literal_substitute(func_type, new_tvars)


def literal_substitute(t: type, type_map: Dict[str, type]) -> type:
    """Make substitutions in t according to type_map, returning resulting type."""
    if isinstance(t, TypeVar) and t.__name__ in type_map:
        return type_map[t.__name__]
    elif isinstance(t, TypeVar):
        return TypeVar(t.__name__)
    elif isinstance(t, ForwardRef):
        return ForwardRef(literal_substitute(t.__forward_arg__, type_map))
    elif isinstance(t, TuplePlus):
        subbed_args = [literal_substitute(t1, type_map) for t1 in t.__constraints__]
        return TuplePlus("tup+", *subbed_args)
    elif is_callable(t):
        args = list(literal_substitute(t1, type_map) for t1 in t.__args__[:-1])
        res = literal_substitute(t.__args__[-1], type_map)
        new_t = Callable[args, res]
        if hasattr(t, "__polymorphic_tvars__"):
            new_t.__polymorphic_tvars__ = t.__polymorphic_tvars__.copy()
        return new_t
    elif isinstance(t, _GenericAlias) and t.__args__ is not None:
        return t.copy_with(tuple(literal_substitute(t1, type_map) for t1 in t.__args__))
    else:
        return t


class Environment:
    """The type bindings for the environment for a particular node.

    Instances of this class contain three dictionaries, representing bindings
    for local, nonlocal, and global bindings.

    TODO: currently, only locals is used; this should be fixed as we add
    the nonlocal and global nodes and use scope information to categorize
    a name binding.
    """

    def __init__(
        self,
        locals_: Optional[Dict[str, type]] = None,
        nonlocals_: Optional[Dict[str, type]] = None,
        globals_: Optional[Dict[str, type]] = None,
    ) -> None:
        """Initialize an environment."""
        self.locals = locals_ or {}
        self.nonlocals = nonlocals_ or {}
        self.globals = globals_ or {}

    def lookup_in_env(self, variable_name: str) -> type:
        """Helper to search for a variable in the environment of a node by name."""
        if variable_name in self.locals:
            return self.locals[variable_name]
        elif variable_name in self.globals:
            return self.globals[variable_name]
        elif variable_name in self.nonlocals:
            return self.nonlocals[variable_name]
        else:
            raise KeyError

    def create_in_env(
        self,
        type_constraints: TypeConstraints,
        environment: "Environment",
        variable_name: str,
        node: nodes.NodeNG,
    ):
        """Helper to create a fresh Type Var and adding the variable to appropriate environment."""
        if environment == "locals":
            self.locals[variable_name] = type_constraints.fresh_tvar(node)
        elif environment == "globals":
            self.globals[variable_name] = type_constraints.fresh_tvar(node)
        elif environment == "nonlocals":
            self.nonlocals[variable_name] = type_constraints.fresh_tvar(node)

    def __str__(self) -> str:
        return str(self.locals)


###############################################################################
# Parsing type annotations
###############################################################################
def parse_annotations(
    node: nodes.NodeNG, class_tvars: Optional[List[type]] = None
) -> List[Tuple[type, str]]:
    """Return types specified by the type annotations for a node.
    Returns more than one type if there are default arguments.
    """
    if isinstance(node, nodes.FunctionDef):
        arg_types = []
        no_class_tvars = class_tvars is None
        is_methodcall = isinstance(node.parent, nodes.ClassDef)
        if no_class_tvars and is_methodcall:
            self_type = _node_to_type(node.parent.name)
        elif no_class_tvars or not is_methodcall:
            self_type = None
        elif node.parent.name in _BUILTIN_TO_TYPING:
            self_type = eval(_BUILTIN_TO_TYPING[node.parent.name])[
                tuple(_node_to_type(tv) for tv in class_tvars)
            ]
        else:
            self_type = _node_to_type(node.parent.name)

        for arg, annotation in zip(node.args.args, node.args.annotations):
            if getattr(arg, "name", None) == "self" and annotation is None:
                arg_types.append(self_type)
            else:
                arg_types.append(_ann_node_to_type(annotation).getValue())

        # Handle optional arguments
        alternatives = []
        for num_optional in range(len(node.args.defaults) + 1):
            alternatives.append(arg_types[: len(arg_types) - num_optional])

        rtype = _ann_node_to_type(node.returns).getValue()

        callables = [
            (create_Callable(arg_types, rtype, class_tvars), node.type)
            for arg_types in alternatives
        ]
        return callables

    elif isinstance(node, nodes.AssignName) and isinstance(node.parent, nodes.AnnAssign):
        return [_ann_node_to_type(node.parent.annotation).getValue(), "attribute"]


def _ann_node_to_type(node: nodes.Name) -> TypeResult:
    """Return a type represented by the input node, substituting Any for missing arguments in generic types"""
    try:
        ann_node_type = _node_to_type(node)
    except SyntaxError:
        # Attempted to create ForwardRef with invalid string
        return TypeFailAnnotationInvalid(node)

    ann_type = _generic_to_annotation(ann_node_type, node)
    return ann_type


def _generic_to_annotation(ann_node_type: type, node: nodes.NodeNG) -> TypeResult:
    is_generic = isinstance(ann_node_type, _GenericAlias) or (
        sys.version_info >= (3, 9) and hasattr(ann_node_type, "__origin__")
    )
    if is_generic and ann_node_type is getattr(
        typing, getattr(ann_node_type, "_name", "") or "", None
    ):
        if ann_node_type == Dict:
            ann_type = wrap_container(ann_node_type, Any, Any)
        elif ann_node_type == Tuple:
            # TODO: Add proper support for multi-parameter Tuples
            ann_type = wrap_container(ann_node_type, Any)
        else:
            ann_type = wrap_container(ann_node_type, Any)
    elif is_generic:
        parsed_args = []
        for arg in ann_node_type.__args__:
            _generic_to_annotation(arg, node) >> parsed_args.append
        ann_type = wrap_container(ann_node_type, *parsed_args)
    else:
        try:
            _type_check(ann_node_type, "")
        except TypeError:
            return TypeFailAnnotationInvalid(node)
        else:
            ann_type = TypeInfo(ann_node_type)
    return ann_type


def _node_to_type(node: nodes.NodeNG, locals: Dict[str, type] = None) -> type:
    """Return a type represented by the input node."""
    locals = locals or _TYPESHED_TVARS
    if node is None:
        return Any
    elif isinstance(node, str):
        return _eval_node(node, globals(), locals)
    elif isinstance(node, nodes.Name):
        return _eval_node(node.name, globals(), locals)
    elif isinstance(node, nodes.Attribute):
        return _eval_node(node.attrname, globals(), locals)
    elif isinstance(node, nodes.Subscript):
        v = _node_to_type(node.value)
        s = _node_to_type(node.slice)
        if isinstance(v, ForwardRef):
            return literal_substitute(v, {v.__forward_arg__: s})
        else:
            return v[s]
    elif isinstance(node, nodes.Index):
        return _node_to_type(node.value)
    elif isinstance(node, nodes.Tuple):
        return tuple(_node_to_type(t) for t in node.elts if not isinstance(t, nodes.Ellipsis))
    elif isinstance(node, nodes.List):
        return [_node_to_type(t) for t in node.elts if not isinstance(t, nodes.Ellipsis)]
    elif isinstance(node, nodes.Const) and node.value is None:
        return None
    elif isinstance(node, nodes.Const) and isinstance(node.value, str):
        return _node_to_type(node.value)
    elif isinstance(node, nodes.Const) and node.value is Ellipsis:
        return Ellipsis
    else:
        return node


def _eval_node(node_name: str, _globals: Dict[str, type], _locals: Dict[str, type]):
    """Return a type represented by node_name."""
    try:
        eval_type = eval(node_name, _globals, _locals)
    except:
        eval_type = ForwardRef(node_name)

    if eval_type in (list, dict, tuple, set):
        # Annotation set as class type (ie. list) instead of typing generic (ie. List[Any])
        return eval(f"typing.{node_name.capitalize()}", _globals, _locals)
    else:
        return eval_type


def _collect_tvars(type: type) -> List[type]:
    if isinstance(type, TypeVar):
        return [type]
    elif isinstance(type, _GenericAlias) and type.__args__:
        return sum([_collect_tvars(arg) for arg in type.__args__], [])
    else:
        return []


def class_callable(init: Callable) -> Callable:
    """Convert an __init__ type signature into a callable for the class."""
    c = init.copy_with(tuple([*init.__args__[1:-1], init.__args__[0]]))
    c.__polymorphic_tvars__ = init.__polymorphic_tvars__.copy()
    return c


def is_callable(t: type) -> bool:
    """Return whether the given type is a typing.Callable type."""
    return getattr(t, "__origin__", None) is Callable.__origin__



"""
Top-level functions to mutate the astroid nodes with `end_col_offset` and
`end_lineno` properties.

Where possible, the `end_col_offset` property is set by that of the node's last child.

    fromlineno
        - existing attribute
        - one-indexed
    end_lineno
        - new attribute
        - one-indexed
    col_offset
        - existing attribute
        - zero-indexed
        - located left of the first character
    end_col_offset
        - new attribute
        - zero-indexed
        - located right of the last character (essentially the string length)

In astroid/astroid/transforms.py, functions are registered to types in the
`transforms` dictionary in the TransformVisitor class. The traversal at
line 83 eventually leads to the transform called on each node at line 36,
within the _transform method.

Astroid Source:
https://github.com/PyCQA/astroid/blob/master/astroid/transforms.py
"""
from astroid import nodes
from astroid.transforms import TransformVisitor

CONSUMABLES = " \n\t\\"


# These nodes have no children, and their end_lineno and end_col_offset
# attributes are set based on their string representation (according to astroid).
# Goal: eventually replace the transforms for all the nodes in this list with the
# predicate technique that uses more robust approach using searching, rather than
# simple length of string.
NODES_WITHOUT_CHILDREN = [
    nodes.AssignName,
    nodes.Break,
    nodes.Const,
    nodes.Continue,
    nodes.DelName,
    nodes.Ellipsis,
    nodes.Global,
    nodes.Import,
    nodes.ImportFrom,
    nodes.Name,
    nodes.Nonlocal,
    nodes.Pass,
    nodes.Yield,
]

# These nodes have a child, and their end_lineno and end_col_offset
# attributes are set equal to those of their last child.
NODES_WITH_CHILDREN = [
    nodes.Assert,
    nodes.Assign,
    nodes.AsyncFor,
    nodes.AsyncFunctionDef,
    nodes.AsyncWith,
    nodes.AugAssign,
    nodes.Await,
    nodes.BinOp,
    nodes.BoolOp,
    nodes.Call,
    nodes.ClassDef,
    nodes.Compare,
    nodes.Comprehension,
    nodes.Decorators,
    nodes.Delete,
    nodes.ExceptHandler,
    nodes.For,
    nodes.FormattedValue,
    nodes.FunctionDef,
    nodes.GeneratorExp,
    nodes.If,
    nodes.IfExp,
    nodes.Index,
    nodes.Keyword,
    nodes.Lambda,
    nodes.List,
    nodes.Module,
    nodes.Raise,
    nodes.Return,
    nodes.Starred,
    nodes.Subscript,
    nodes.TryExcept,
    nodes.TryFinally,
    nodes.UnaryOp,
    nodes.While,
    nodes.With,
    nodes.YieldFrom,
]


# Predicate functions, for setting locations based on source code.
# Predicates can only return a single truthy value, because of how its used in
# `astroid/transforms.py`
# ====================================================
def _token_search(token):
    """
    @type token: string
    @rtype: function
    """

    def _is_token(s, index, node):
        """Fix to include certain tokens such as a paren, bracket, or brace.
        @type s: string
        @type index: int
        @type node: Astroid node
        @rtype: bool
        """
        return s[index] == token

    return _is_token


def _keyword_search(keyword):
    """
    @type keyword: string
    @rtype: function
    """

    def _is_keyword(s, index, node):
        """Search for a keyword. Right-to-left.
        @type s: string
        @type index: int
        @type node: Astroid node
        @rtype: bool
        """
        return s[index : index + len(keyword)] == keyword

    return _is_keyword


def _is_attr_name(s, index, node):
    """Search for the name of the attribute. Left-to-right."""
    target_len = len(node.attrname)
    if index < target_len:
        return False
    return s[index - target_len + 1 : index + 1] == node.attrname


def _is_arg_name(s, index, node):
    """Search for the name of the argument. Right-to-left."""
    if not node.arg:
        return False
    return s[index : index + len(node.arg)] == node.arg


# Nodes the require the source code for proper location setting
# Elements here are in the form
# (node class, predicate for start | None, predicate for end | None)
NODES_REQUIRING_SOURCE = [
    (nodes.AssignAttr, None, _is_attr_name),
    (nodes.AsyncFor, _keyword_search("async"), None),
    (nodes.AsyncFunctionDef, _keyword_search("async"), None),
    (nodes.AsyncWith, _keyword_search("async"), None),
    (nodes.Attribute, None, _is_attr_name),
    (nodes.Call, None, _token_search(")")),
    (nodes.DelAttr, _keyword_search("del"), _is_attr_name),
    (nodes.DelName, _keyword_search("del"), None),
    (nodes.Dict, None, _token_search("}")),
    (nodes.DictComp, None, _token_search("}")),
    (nodes.Expr, _token_search("("), _token_search(")")),
    # TODO: use same behavior as Slice.
    (nodes.ExtSlice, _token_search("["), _token_search("]")),
    (nodes.GeneratorExp, _token_search("("), _token_search(")")),
    (nodes.If, _keyword_search("elif"), None),
    (nodes.Index, _token_search("["), _token_search("]")),
    (nodes.Keyword, _is_arg_name, None),
    (nodes.List, _token_search("["), _token_search("]")),
    (nodes.ListComp, _token_search("["), _token_search("]")),
    (nodes.Set, None, _token_search("}")),
    (nodes.SetComp, None, _token_search("}")),
    (nodes.Slice, _token_search("["), None),
    (nodes.Tuple, None, _token_search(",")),
]


def init_register_ending_setters(source_code):
    """Instantiate a visitor to transform the nodes.
    Register the transform functions on an instance of TransformVisitor.

    @type source_code: list of strings
    @rtype: TransformVisitor
    """
    ending_transformer = TransformVisitor()

    # Check consistency of astroid-provided fromlineno and col_offset attributes.
    for node_class in nodes.ALL_NODE_CLASSES:
        ending_transformer.register_transform(
            node_class,
            fix_start_attributes,
            lambda node: (
                getattr(node, "fromlineno", None) is None
                or getattr(node, "col_offset", None) is None
            ),
        )

    # Ad hoc transformations
    ending_transformer.register_transform(nodes.BinOp, _set_start_from_first_child)
    ending_transformer.register_transform(nodes.ClassDef, _set_start_from_first_child)
    ending_transformer.register_transform(nodes.FunctionDef, _set_start_from_first_child)
    ending_transformer.register_transform(nodes.Tuple, _set_start_from_first_child)
    ending_transformer.register_transform(nodes.Arguments, fix_arguments(source_code))
    ending_transformer.register_transform(nodes.Slice, fix_slice(source_code))

    for node_class in NODES_WITHOUT_CHILDREN:
        ending_transformer.register_transform(node_class, set_without_children)
    for node_class in NODES_WITH_CHILDREN:
        ending_transformer.register_transform(node_class, set_from_last_child)

    ending_transformer.register_transform(nodes.Subscript, fix_subscript(source_code))

    # Nodes where the source code must also be provided.
    # source_code and the predicate functions get stored in the TransformVisitor
    for node_class, start_pred, end_pred in NODES_REQUIRING_SOURCE:
        if start_pred is not None:
            ending_transformer.register_transform(
                node_class, start_setter_from_source(source_code, start_pred)
            )
        if end_pred is not None:
            # This is for searching for a trailing comma after a tuple's final element
            if node_class is nodes.Tuple:
                ending_transformer.register_transform(
                    node_class, end_setter_from_source(source_code, end_pred, True)
                )
            else:
                ending_transformer.register_transform(
                    node_class, end_setter_from_source(source_code, end_pred)
                )

    # Nodes where extra parentheses are included
    ending_transformer.register_transform(nodes.BinOp, add_parens(source_code))
    ending_transformer.register_transform(nodes.Const, add_parens(source_code))
    ending_transformer.register_transform(nodes.Tuple, add_parens(source_code))

    return ending_transformer


# Transform functions.
# These functions are called on individual nodes to either fix the
# `fromlineno` and `col_offset` properties of the nodes,
# or to set the `end_lineno` and `end_col_offset` attributes for a node.
# ====================================================
def fix_slice(source_code):
    """
    The Slice node column positions are mostly set properly when it has (Const)
    children. The main problem is when Slice node doesn't have children.
    E.g "[:]", "[::]", "[:][:]", "[::][::]", ... yikes! The existing positions
    are sometimes set improperly to 0.
    """

    def _find_square_brackets(node):
        if _get_last_child(node):
            set_from_last_child(node)
            line_i = node.end_lineno - 1  # convert 1 to 0 index.
            char_i = node.end_col_offset
            has_children = True
        else:
            line_i = node.parent.value.end_lineno - 1  # convert 1 to 0 index.
            char_i = node.parent.value.end_col_offset
            has_children = False

        # Search the remaining source code for the "]" char.
        while char_i < len(source_code[line_i]) and source_code[line_i][char_i] != "]":
            if char_i == len(source_code[line_i]) - 1 or source_code[line_i][char_i] == "#":
                char_i = 0
                line_i += 1
            else:
                char_i += 1

        if not has_children:
            node.fromlineno, node.col_offset = line_i + 1, char_i
        node.end_lineno, node.end_col_offset = line_i + 1, char_i + 1

        return node

    return _find_square_brackets


def fix_subscript(source_code):
    """For a Subscript node.

    Need to include this because the index/extended slice is a value rather than
    a separate Index/ExtSlice in Python 3.9.
    """

    def _fix_end(node: nodes.Subscript) -> nodes.Subscript:
        if isinstance(node.slice, nodes.Slice):
            # In this case, the subscript node already contains the final ].
            return node

        # Search the remaining source code for the "]" char.
        if _get_last_child(node):
            set_from_last_child(node)
            line_i = node.end_lineno - 1  # convert 1 to 0 index.
            char_i = node.end_col_offset
        else:
            line_i = node.value.end_lineno - 1  # convert 1 to 0 index.
            char_i = node.value.end_col_offset

        while char_i < len(source_code[line_i]) and source_code[line_i][char_i] != "]":
            if char_i == len(source_code[line_i]) - 1 or source_code[line_i][char_i] == "#":
                char_i = 0
                line_i += 1
            else:
                char_i += 1

        node.end_lineno, node.end_col_offset = line_i + 1, char_i + 1
        return node

    return _fix_end


def fix_arguments(source_code):
    """For an Arguments node"""

    def _find(node: nodes.Arguments) -> nodes.Arguments:
        children = list(node.get_children())
        if children:
            fix_start_attributes(node)

        line_i = node.parent.fromlineno
        char_i = node.parent.col_offset
        for child in children:
            if line_i is None:
                line_i = child.end_lineno
                char_i = child.end_col_offset
            elif (
                line_i < child.end_lineno
                or line_i == child.end_lineno
                and char_i < child.end_col_offset
            ):
                line_i = child.end_lineno
                char_i = child.end_col_offset

        line_i -= 1  # Switch to 0-indexing

        # left bracket if parent is FunctionDef, colon if Lambda
        if isinstance(node.parent, nodes.FunctionDef):
            end_char = ")"
        else:
            end_char = ":"

        while char_i < len(source_code[line_i]) and source_code[line_i][char_i] != end_char:
            if char_i == len(source_code[line_i]) - 1 or source_code[line_i][char_i] == "#":
                char_i = 0
                line_i += 1
            else:
                char_i += 1

        node.end_lineno, node.end_col_offset = line_i + 1, char_i

        # no children
        if children == []:
            node.fromlineno, node.col_offset = line_i + 1, char_i

        return node

    return _find


def fix_start_attributes(node):
    """Some nodes don't always have the `col_offset` property set by Astroid:
    Comprehension, ExtSlice, Index, Keyword, Module, Slice.
    """
    try:
        first_child = next(node.get_children())
        if getattr(node, "fromlineno", None) is None:
            node.fromlineno = first_child.fromlineno
        if getattr(node, "col_offset", None) is None:
            node.col_offset = first_child.col_offset

    except StopIteration:
        # No children. Go to the enclosing statement and use that.
        # This assumes that statement nodes will always have these attributes set.
        statement = node.statement()
        if statement is not node:
            if getattr(node, "fromlineno", None) is None:
                node.fromlineno = statement.fromlineno
            if getattr(node, "col_offset", None) is None:
                node.col_offset = statement.col_offset
        else:
            # Enclosing statement is same as node, also does not have attributes set
            if getattr(node, "fromlineno", None) is None:
                node.fromlineno = 0
            if getattr(node, "col_offset", None) is None:
                node.col_offset = 0
    return node


def _set_start_from_first_child(node):
    """Set the start attributes of this node from its first child."""
    try:
        first_child = next(node.get_children())
    except StopIteration:
        pass
    else:
        node.fromlineno = first_child.fromlineno
        node.col_offset = first_child.col_offset
    return node


def set_from_last_child(node):
    """Populate ending locations for astroid node based on its last child.

    Preconditions:
      - `node` must have a `last_child` (node).
      - `node` has col_offset property set.
    """
    last_child = _get_last_child(node)
    if not last_child:
        set_without_children(node)
        return node
    elif not hasattr(last_child, "end_lineno"):  # Newly added for Slice() node.
        set_without_children(last_child)

    if last_child.end_lineno is not None:
        node.end_lineno = last_child.end_lineno
    if last_child.end_col_offset is not None:
        node.end_col_offset = last_child.end_col_offset
    return node


def set_without_children(node):
    """Populate ending locations for nodes that are guaranteed to never have
    children. E.g. Const.

    These node's end_col_offset are currently assigned based on their
    computed string representation. This may differ from their actual
    source code representation, however (mainly whitespace).

    Precondition: `node` must not have a `last_child` (node).
    """
    if not hasattr(node, "end_lineno"):
        node.end_lineno = node.fromlineno
    # FIXME: using the as_string() is a bad technique because many different
    # whitespace possibilities that may not be reflected in it!
    if not hasattr(node, "end_col_offset"):
        node.end_col_offset = node.col_offset + len(node.as_string())
    return node


def _get_last_child(node):
    """Returns the last child node, or None.
    Some nodes' last_child() attribute not set, e.g. nodes.Arguments.
    """
    if node.last_child():
        return node.last_child()
    else:
        # Get the first child from the `get_children` generator.
        skip_to_last_child = None  # save reference to last child.
        for skip_to_last_child in node.get_children():
            pass  # skip to last
        return skip_to_last_child  # postcondition: node, or None.


def end_setter_from_source(source_code, pred, only_consumables=False):
    """Returns a *function* that sets ending locations for a node from source.

    The basic technique is to do the following:
      1. Find the ending locations for the node based on its last child.
      2. Starting at that point, iterate through characters in the source code
         up to and including the first index that satisfies pred.

    pred is a function that takes a string and index and returns a bool,
    e.g. _is_close_paren

    If only_consumables is True, the search halts when it reaches a non-consumable
    character that fails pred *on the first line*.
    TODO: really the behaviour should be the same for all lines searched for.
    """

    def set_endings_from_source(node):
        if not hasattr(node, "end_col_offset"):
            set_from_last_child(node)

        # Initialize counters. Note: we need to offset lineno,
        # since it's 1-indexed.
        end_col_offset, lineno = node.end_col_offset, node.end_lineno - 1

        # First, search the remaining part of the current end line.
        for j in range(end_col_offset, len(source_code[lineno])):
            if source_code[lineno][j] == "#":
                break  # skip over comment lines
            if pred(source_code[lineno], j, node):
                node.end_col_offset = j + 1
                return node
            elif only_consumables and source_code[lineno][j] not in CONSUMABLES:
                return node

        # If that doesn't work, search remaining lines
        for i in range(lineno + 1, len(source_code)):
            # Search each character
            for j in range(len(source_code[i])):
                if source_code[i][j] == "#":
                    break  # skip over comment lines
                if pred(source_code[i], j, node):
                    node.end_col_offset, node.end_lineno = j + 1, i + 1
                    return node
                # only consume inert characters.
                elif source_code[i][j] not in CONSUMABLES:
                    return node
        return node

    return set_endings_from_source


def start_setter_from_source(source_code, pred):
    """Returns a *function* that sets start locations for a node from source.
    Recall `source_code`, `pred` are within the lexical scope of the returned function.

    The basic technique is to do the following:
      1. Find the start locations for the node (already set).
      2. Starting at that point, iterate through characters in the source code
         in reverse until reaching the first index that satisfies pred.

    pred is a function that takes a string and index and returns a bool,
    e.g. _is_open_paren
    """

    def set_start_from_source(node):
        # Initialize counters. Note: fromlineno is 1-indexed.
        col_offset, lineno = node.col_offset, node.fromlineno - 1

        # First, search the remaining part of the current start line
        for j in range(min(len(source_code[lineno]) - 1, col_offset), -1, -1):
            if pred(source_code[lineno], j, node):
                node.col_offset = j
                return node

        # If that doesn't work, search remaining lines
        for i in range(lineno - 1, -1, -1):
            # Search each character, right-to-left
            for j in range(len(source_code[i]) - 1, -1, -1):
                if pred(source_code[i], j, node):
                    node.end_col_offset, node.end_lineno = j, i + 1
                    return node
                # only consume inert characters.
                elif source_code[i][j] not in CONSUMABLES:
                    return node
        return node

    return set_start_from_source


def add_parens(source_code):
    def h(node):
        _add_parens(source_code)(node)

    return h


def _add_parens(source_code):
    def h(node):
        # Initialize counters. Note: fromlineno is 1-indexed.
        prev = node.fromlineno, node.col_offset, node.end_lineno, node.end_col_offset
        while True:
            col_offset, lineno = node.col_offset, node.fromlineno - 1
            end_col_offset, end_lineno = node.end_col_offset, node.end_lineno - 1

            # First, search the remaining part of the current start line
            prev_char, new_lineno, new_coloffset = None, None, None
            for j in range(col_offset - 1, -1, -1):
                if source_code[lineno][j] in CONSUMABLES or source_code[lineno][j] == ",":
                    continue
                else:
                    prev_char, new_lineno, new_coloffset = source_code[lineno][j], lineno, j
                    break

            if prev_char is None:
                # Search remaining lines
                for i in range(lineno - 1, -1, -1):
                    # Search each character, right-to-left
                    for j in range(len(source_code[i]) - 1, -1, -1):
                        if source_code[i][j] in CONSUMABLES or source_code[i][j] == ",":
                            continue
                        else:
                            prev_char, new_lineno, new_coloffset = source_code[i][j], i, j

                            break
                    if prev_char is not None:
                        break

            if prev_char != "(":
                # No enclosing parentheses
                break

            # Now search for matching ')'
            next_char, new_end_lineno, new_end_coloffset = None, None, None
            for j in range(end_col_offset, len(source_code[end_lineno])):
                if source_code[end_lineno][j] == "#":
                    break  # skip over comment lines
                elif source_code[end_lineno][j] in CONSUMABLES:
                    continue
                else:
                    next_char, new_end_lineno, new_end_coloffset = (
                        source_code[end_lineno][j],
                        end_lineno,
                        j,
                    )
                    break

            if next_char is None:
                # Search remaining lines
                for i in range(end_lineno + 1, len(source_code)):
                    # Search each character
                    for j in range(len(source_code[i])):
                        if source_code[i][j] == "#":
                            break  # skip over comment lines
                        elif source_code[i][j] in CONSUMABLES:
                            continue
                        else:
                            next_char, new_end_lineno, new_end_coloffset = source_code[i][j], i, j
                            break
                    if next_char is not None:
                        break

            if next_char != ")":
                break

            # At this point, an enclosing pair of parentheses has been found
            prev = node.fromlineno, node.col_offset, node.end_lineno, node.end_col_offset
            node.fromlineno, node.col_offset, node.end_lineno, node.end_col_offset = (
                new_lineno + 1,
                new_coloffset,
                new_end_lineno + 1,
                new_end_coloffset + 1,
            )

        # Go back by 1 set of parentheses if inside a function call.
        if isinstance(node.parent, nodes.Call) and len(node.parent.args) == 1:
            node.fromlineno, node.col_offset, node.end_lineno, node.end_col_offset = prev

        return node

    return h


# Make this module a pylint plugin
def register(linter):
    old_get_ast = linter.get_ast

    def new_get_ast(filepath, modname):
        ast = old_get_ast(filepath, modname)
        with open(filepath) as f:
            source_code = f.readlines()
        ending_transformer = TransformVisitor()
        register_transforms(source_code, ending_transformer)
        ending_transformer.visit(ast)
        return ast

    linter.get_ast = new_get_ast


def register_transforms(source_code, obj):
    # Check consistency of astroid-provided fromlineno and col_offset attributes.
    for node_class in nodes.ALL_NODE_CLASSES:
        obj.register_transform(
            node_class,
            fix_start_attributes,
            lambda node: (
                getattr(node, "fromlineno", None) is None
                or getattr(node, "col_offset", None) is None
            ),
        )

    # Ad hoc transformations
    obj.register_transform(nodes.Arguments, fix_arguments(source_code))

    for node_class in NODES_WITH_CHILDREN:
        obj.register_transform(node_class, set_from_last_child)
    for node_class in NODES_WITHOUT_CHILDREN:
        obj.register_transform(node_class, set_without_children)

    # Nodes where the source code must also be provided.
    for node_class, start_pred, end_pred in NODES_REQUIRING_SOURCE:
        if start_pred is not None:
            obj.register_transform(node_class, start_setter_from_source(source_code, start_pred))
        if end_pred is not None:
            obj.register_transform(node_class, end_setter_from_source(source_code, end_pred))



import sys
from typing import List, Optional

import click

from python_ta import check_all, check_errors

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="python_ta configuration file",
)
@click.option("-E", "--errors-only", is_flag=True, help="Displays errors only", default=False)
@click.argument(
    "filenames", nargs=-1, type=click.Path(exists=True, dir_okay=True, resolve_path=True)
)
@click.option("--exit-zero", is_flag=True, help="Always return with status code 0", default=False)
def main(config: Optional[str], errors_only: bool, filenames: List[str], exit_zero: bool) -> None:
    """A code checking tool for teaching Python.

    FILENAMES can be a string of a directory, or file to check (`.py` extension optional) or
    a list of strings of directories or files.
    """
    # `config` is None if `-c` flag is not set
    checker = check_errors if errors_only else check_all
    paths = [click.format_filename(fn) for fn in filenames]
    reporter = checker(module_name=paths, config=config)

    if not exit_zero and reporter.has_messages():
        sys.exit(1)
    else:
        sys.exit(0)


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Generator, List, Optional, Set, Tuple, Union

from astroid import Break, Continue, NodeNG, Return


class ControlFlowGraph:
    """A graph representing the control flow of a Python program."""

    start: CFGBlock
    end: CFGBlock
    # block_count is used as an "autoincrement" to ensure the block ids are unique.
    block_count: int
    # blocks (with at least one statement) that will never be executed in runtime.
    unreachable_blocks: Set[CFGBlock]

    def __init__(self) -> None:
        self.block_count = 0
        self.unreachable_blocks = set()
        self.start = self.create_block()
        self.end = self.create_block()

    def create_block(self, pred: Optional[CFGBlock] = None) -> CFGBlock:
        """Create a new CFGBlock for this graph.

        If pred is specified, set that block as a predecessor of the new block.
        """
        new_block = CFGBlock(self.block_count)
        self.unreachable_blocks.add(new_block)

        self.block_count += 1
        if pred:
            self.link_or_merge(pred, new_block)
        return new_block

    def link(self, source: CFGBlock, target: CFGBlock) -> None:
        """Link source to target."""
        if not source.is_jump():
            CFGEdge(source, target)

    def link_or_merge(self, source: CFGBlock, target: CFGBlock) -> None:
        """Link source to target, or merge source into target if source is empty.

        An "empty" node for this purpose is when source has no statements.

        source with a jump statement cannot be further linked or merged to
        another target.
        """
        if source.is_jump():
            return
        if source.statements == []:
            if source is self.start:
                self.start = target
            else:
                for edge in source.predecessors:
                    edge.target = target
                    target.predecessors.append(edge)
            # source is a utility block that helps build the cfg that does not
            # represent any part of the program so it is redundant.
            self.unreachable_blocks.remove(source)
        else:
            CFGEdge(source, target)

    def multiple_link_or_merge(self, source: CFGBlock, targets: List[CFGBlock]) -> None:
        """Link source to multiple target, or merge source into targets if source is empty.

        An "empty" node for this purpose is when source has no statements.

        source with a jump statement cannot be further linked or merged to
        another target.

        Precondition:
            - source != cfg.start
        """
        if source.statements == []:
            for edge in source.predecessors:
                for t in targets:
                    CFGEdge(edge.source, t)
                edge.source.successors.remove(edge)
            source.predecessors = []
            self.unreachable_blocks.remove(source)
        else:
            for target in targets:
                self.link(source, target)

    def get_blocks(self) -> Generator[CFGBlock, None, None]:
        """Generate a sequence of all blocks in this graph."""
        yield from self._get_blocks(self.start, set())

    def _get_blocks(self, block: CFGBlock, visited: Set[int]) -> Generator[CFGBlock, None, None]:
        if block.id in visited:
            return

        yield block
        visited.add(block.id)

        for edge in block.successors:
            yield from self._get_blocks(edge.target, visited)

    def get_blocks_postorder(self) -> Generator[CFGBlock, None, None]:
        """Return the sequence of all blocks in this graph in the order of
        a post-order traversal."""
        yield from self._get_blocks_postorder(self.start, set())

    def _get_blocks_postorder(self, block: CFGBlock, visited) -> Generator[CFGBlock, None, None]:
        if block.id in visited:
            return

        visited.add(block.id)
        for succ in block.successors:
            yield from self._get_blocks_postorder(succ.target, visited)

        yield block

    def get_edges(self) -> Generator[CFGEdge, None, None]:
        """Generate a sequence of all edges in this graph."""
        yield from self._get_edges(self.start, set())

    def _get_edges(self, block: CFGBlock, visited: Set[int]) -> Generator[CFGEdge, None, None]:
        if block.id in visited:
            return

        visited.add(block.id)

        for edge in block.successors:
            yield edge
            yield from self._get_edges(edge.target, visited)

    def update_block_reachability(self) -> None:
        for block in self.get_blocks():
            block.reachable = True
            if block in self.unreachable_blocks:
                self.unreachable_blocks.remove(block)


class CFGBlock:
    """A node in a control flow graph.

    Represents a maximal block of code whose statements are guaranteed to execute in sequence.
    """

    # A unique identifier
    id: int
    # The statements in this block.
    statements: List[NodeNG]
    # This block's in-edges (from blocks that can execute immediately before this one).
    predecessors: List[CFGEdge]
    # This block's out-edges (to blocks that can execute immediately after this one).
    successors: List[CFGEdge]
    # Whether there exists a path from the start block to this block.
    reachable: bool

    def __init__(self, id_: int) -> None:
        """Initialize a new CFGBlock."""
        self.id = id_
        self.statements = []
        self.predecessors = []
        self.successors = []
        self.reachable = False

    def add_statement(self, statement: NodeNG) -> None:
        if not self.is_jump():
            self.statements.append(statement)
            statement.cfg_block = self

    @property
    def jump(self) -> Optional[NodeNG]:
        if len(self.statements) > 0:
            return self.statements[-1]

    def is_jump(self) -> bool:
        """Returns True if the block has a statement that branches
        the control flow (ex: `break`)"""
        return isinstance(self.jump, (Break, Continue, Return))


class CFGEdge:
    """An edge in a control flow graph.

    Edges are directed, and in the future may be augmented with auxiliary metadata about the control flow.
    """

    source: CFGBlock
    target: CFGBlock

    def __init__(self, source: CFGBlock, target: CFGBlock) -> None:
        self.source = source
        self.target = target
        self.source.successors.append(self)
        self.target.predecessors.append(self)



import sys
import typing
from typing import ForwardRef, _GenericAlias


def _get_name(t: type) -> str:
    """If t is associated with a class, return the name of the class; otherwise, return a string repr. of t"""
    if isinstance(t, ForwardRef):
        return t.__forward_arg__
    elif isinstance(t, type):
        return t.__name__
    elif isinstance(t, _GenericAlias):
        return "{} of {}".format(
            _get_name(t.__origin__), ", ".join(_get_name(arg) for arg in t.__args__)
        )
    else:
        return str(t)


def _gorg(x):
    """Make _gorg compatible for Python 3.6.2 and 3.6.3."""
    if sys.version_info >= (3, 7, 0):
        return x.__origin__
    if sys.version_info < (3, 6, 3):
        return typing._gorg(x)
    else:
        return x._gorg



from .plain_reporter import PlainReporter


class StatReporter(PlainReporter):

    error_messages = []
    style_messages = []

    def __init__(self, source_lines=None):
        """Initialize a StatReporter.

        Clear the two class-level message lists.

        @type source_lines: List[str]
        @rtype: None
        """
        super().__init__(source_lines)
        StatReporter.error_messages = []
        StatReporter.style_messages = []

    def print_messages(self, level="all"):
        """Override the corresponding function in PlainReporter.

        @type level: str
        @rtype: None
        """
        StatReporter.error_messages.extend(self._error_messages)
        StatReporter.style_messages.extend(self._style_messages)

    # to appease PyCharm's NotImplemented complaint
    _display = None



from .color_reporter import ColorReporter
from .html_reporter import HTMLReporter
from .json_reporter import JSONReporter
from .plain_reporter import PlainReporter

# Export tuple of reporter classes for python_ta init file.
REPORTERS = (ColorReporter, PlainReporter, HTMLReporter, JSONReporter)



from astroid import nodes
from hypothesis import HealthCheck, given, settings

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


@given(cs.simple_homogeneous_dict_node(min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_dict_comprehension_reproduce_homogeneous(node):
    """Test DictComp node visitor representing a comprehension expression reproducing the the key, value pairs
    of a homogeneous Dictionary."""
    dictionary = node.as_string()
    program = f"{{key: {dictionary}[key] for key in {dictionary}}}"
    module, _ = cs._parse_text(program)
    dictcomp_node = list(module.nodes_of_class(nodes.DictComp))[0]
    # reproducing the dict with dict. comprehension; the type of the expression will be same as original iterable
    assert dictcomp_node.inf_type.getValue() == dictcomp_node.generators[0].iter.inf_type.getValue()


@given(cs.dict_node(min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_dict_comprehension_reproduce_heterogeneous(node):
    """Test DictComp node visitor representing a comprehension expression reproducing the the key, value pairs
    of a heterogeneous Dictionary."""
    dictionary = node.as_string()
    program = f"{{key: {dictionary}[key] for key in {dictionary}}}"
    module, _ = cs._parse_text(program)
    dictcomp_node = list(module.nodes_of_class(nodes.DictComp))[0]
    assert dictcomp_node.inf_type.getValue() == dictcomp_node.generators[0].iter.inf_type.getValue()



"""Run from the `pyta` root directory to use the local `python_ta` rather than
installed `python_ta` package.
"""
import subprocess
import sys
from os import path, remove
from unittest.mock import Mock

from pytest import skip

import python_ta


def test_check_on_dir():
    """The function, check_all() should handle a top-level dir as input."""
    _inputs = [["nodes"], ["examples"]]
    for item in _inputs:
        python_ta.check_all(
            item,
            config={
                "output-format": "python_ta.reporters.PlainReporter",
                "pyta-error-permission": "no",
                "pyta-file-permission": "no",
            },
        )


def test_check_on_file():
    """Test files"""
    _inputs = [["nodes/name.py"], ["nodes/dict.py", "nodes/const.py"]]
    for item in _inputs:
        python_ta.check_all(
            item,
            config={
                "output-format": "python_ta.reporters.PlainReporter",
                "pyta-error-permission": "no",
                "pyta-file-permission": "no",
            },
        )


def test_check_on_bad_input():
    """Test bad inputs. In all cases, pyta should recover.
    Any valid files, if any, should be checked.
    """
    _inputs = [
        [222],
        222,
        ["nodes/dict.py nodes/const.py"],
        [222, "examples/inline_config_comment.py", "nodes/dict.py"],
        ["file_does_not_exist"],
    ]
    for item in _inputs:
        python_ta.check_all(
            item,
            config={
                "output-format": "python_ta.reporters.PlainReporter",
                "pyta-error-permission": "no",
                "pyta-file-permission": "no",
            },
        )


def test_check_with_config():
    """Test inputs along with a config arg."""
    _inputs = [["nodes/const.py"], ["nodes"]]
    CONFIG = {
        # [ELIF]
        "max-nested-blocks": 4,
        # [FORMAT]
        "max-line-length": 80,
        # [FORBIDDEN IMPORT]
        "allowed-import-modules": ["doctest", "unittest", "hypothesis", "python_ta"],
        # [FORBIDDEN IO]
        "allowed-io": [],
        # [MESSAGES CONTROL]
        "disable": [
            "R0401",
            "R0901",
            "R0903",
            "R0904",
            "R0911",
            "R0916",
            "W0402",
            "W0403",
            "W0410",
            "W1501",
            "W1502",
            "W1505",
            "E1300",
            "E1301",
            "E1302",
            "E1304",
            "W1300",
            "W1301",
            "W1302",
            "W1304",
            "E1124",
            "E1125",
            "E1129",
            "E1132",
            "W1402",
            "W0105",
            "E1303",
            "W1306",
            "W1307",
            "E0116",
            "E0114",
            "E0112",
            "E0115",
            "E0106",
            "E0113",
            "E0111",
            "E0105",
            "E0100",
            "E0117",
            "W0150",
            "W0120",
            "W0124",
            "W0108",
            "W0123",
            "W0122",
            "W0110",
            "C0122",
            "C0200",
            "W0141",
            "W0640",
            "W0623",
            "W0614",
            "W0604",
            "W0603",
            "W0602",
            "W0601",
            "E0604",
            "E0603",
            "E1200",
            "E1201",
            "E1202",
            "W1201",
            "E1205",
            "E1206",
            "similarities",
            "newstyle",
            "python3",
            "W0512",
            "C0403",
            "C0401",
            "C0402",
            "E1701",
            "E1700",
            "W0332",
            "C0327",
            "C0328",
            "E0202",
            "E0241",
            "E0704",
            "W0211",
            "W0232",
            "W0511",
            "R0204",
            "C0303",
            "W0231",
        ],
        # [CUSTOM PYTA OPTIONS]
        "output-format": "python_ta.reporters.PlainReporter",
        "pyta-error-permission": "no",
        "pyta-file-permission": "no",
    }
    for item in _inputs:
        python_ta.check_all(item, config=CONFIG)


def test_check_saves_file() -> None:
    """Test whether or not specifiying an output properly saves a file"""
    _inputs = [["nodes/name.py"]]
    for item in _inputs:
        # Note that the reporter output will be created in the main directory
        python_ta.check_all(item, output="pyta_output.html")

    file_exists = path.exists("pyta_output.html")

    assert file_exists

    # If the file exists, the assertion passes and the file gets removed from the main directory
    if file_exists:
        remove("pyta_output.html")


def test_check_no_reporter_output() -> None:
    """Test whether not specifiying an output does not save a file"""
    # To prevent any any browser/server code running when running Pytest to avoid CI timeouts
    import webbrowser
    from http.server import HTTPServer

    webbrowser.open = Mock(return_value=None)
    HTTPServer.handle_request = Mock(return_value=None)

    _inputs = [["nodes/name.py"]]
    for item in _inputs:
        # Note that the reporter output *would have been* created in the main directory
        python_ta.check_all(item)

    file_exists = path.exists("pyta_output.html")

    assert not file_exists

    # If the file exists, the assertion failed and the file gets removed from main directory
    if file_exists:
        remove("pyta_output.html")


def test_check_no_errors_zero() -> None:
    """Test that python_ta exits with status code 0 when it does not detect errors."""
    output = subprocess.run(
        [
            sys.executable,
            "-m",
            "python_ta",
            "--config",
            "tests/test.pylintrc",
            "tests/fixtures/no_errors.py",
        ]
    )

    assert output.returncode == 0


def test_check_errors_nonzero() -> None:
    """Test that python_ta exits with non-zero status code when it detects errors."""
    output = subprocess.run(
        [
            sys.executable,
            "-m",
            "python_ta",
            "--config",
            "tests/test.pylintrc",
            "nodes/name.py",
        ]
    )

    assert output.returncode != 0


def test_check_exit_zero() -> None:
    """Test that python_ta --exit-zero always exits with status code 0,
    even when given a file with errors.
    """
    output = subprocess.run(
        [
            sys.executable,
            "-m",
            "python_ta",
            "--exit-zero",
            "--config",
            "tests/test.pylintrc",
            "nodes/name.py",
        ]
    )

    assert output.returncode == 0



"""Python Teaching Assistant

The goal of this module is to provide automated feedback to students in our
introductory Python courses, using static analysis of their code.

To run the checker, call the check function on the name of the module to check.

> import python_ta
> python_ta.check_all('mymodule.py')

Or, put the following code in your Python module:

if __name__ == '__main__':
    import python_ta
    python_ta.check_all()
"""
__version__ = "2.1.1"  # Version number

# First, remove underscore from builtins if it has been bound in the REPL.
import builtins

try:
    del builtins._
except AttributeError:
    pass

import importlib.util
import os
import sys
import tokenize
import webbrowser
from typing import Generator

import pylint.config
import pylint.lint
import pylint.utils
from astroid import MANAGER, modutils
from pylint.utils.pragma_parser import OPTION_PO

from .patches import patch_all
from .reporters import REPORTERS
from .upload import upload_to_server

HELP_URL = "http://www.cs.toronto.edu/~david/pyta/"

# check the python version
if sys.version_info < (3, 7, 0):
    print("[WARNING] You need Python 3.7 or later to run PythonTA.")


# Flag to determine if we've previously patched pylint
PYLINT_PATCHED = False


def check_errors(module_name="", config="", output=None):
    """Check a module for errors, printing a report."""
    return _check(module_name=module_name, level="error", local_config=config, output=output)


def check_all(module_name="", config="", output=None):
    """Check a module for errors and style warnings, printing a report."""
    return _check(module_name=module_name, level="all", local_config=config, output=output)


def _check(module_name="", level="all", local_config="", output=None):
    """Check a module for problems, printing a report.

    The `module_name` can take several inputs:
      - string of a directory, or file to check (`.py` extension optional).
      - list of strings of directories or files -- can have multiple.
      - no argument -- checks the python file containing the function call.
    `level` is used to specify which checks should be made.
    `local_config` is a dict of config options or string (config file name).
    `output` is an absolute or relative path to capture pyta data output. Default std out.
    """
    # Manually create pylint cache directory if it doesn't exist.
    # This is a backport of https://github.com/PyCQA/pylint/pull/4988,
    # and should be removed once we update to pylint 2.11.
    if not os.path.exists(pylint.config.PYLINT_HOME):
        try:
            os.makedirs(pylint.config.PYLINT_HOME)
        except OSError:
            pass

    linter = reset_linter(config=local_config)
    current_reporter = linter.reporter
    current_reporter.set_output(output)

    global PYLINT_PATCHED
    if not PYLINT_PATCHED:
        patch_all()  # Monkeypatch pylint (override certain methods)
        PYLINT_PATCHED = True

    # Try to check file, issue error message for invalid files.
    try:
        for locations in _get_valid_files_to_check(module_name):
            f_paths = []  # Paths to files for data submission
            errs = []  # Errors caught in files for data submission
            config = {}  # Configuration settings for data submission
            for file_py in get_file_paths(locations):
                if not _verify_pre_check(file_py):
                    continue  # Check the other files
                # Load config file in user location. Construct new linter each
                # time, so config options don't bleed to unintended files.
                # Reuse the same reporter each time to accumulate the results across different files.
                linter = reset_linter(config=local_config, file_linted=file_py)
                linter.set_reporter(current_reporter)
                module_name = os.path.splitext(os.path.basename(file_py))[0]
                if module_name in MANAGER.astroid_cache:  # Remove module from astroid cache
                    del MANAGER.astroid_cache[module_name]
                linter.check(file_py)  # Lint !
                current_reporter.print_messages(level)
                if linter.config.pyta_file_permission:
                    f_paths.append(file_py)  # Appending paths for upload
                print(
                    "[INFO] File: {} was checked using the configuration file: {}".format(
                        file_py, linter.config_file
                    ),
                    file=sys.stderr,
                )
            if linter.config.pyta_error_permission:
                errs = list(current_reporter.messages.values())
            if (
                f_paths != [] or errs != []
            ):  # Only call upload_to_server() if there's something to upload
                # Checks if default configuration was used without changing options through the local_config argument
                if linter.config_file[-19:-10] != "python_ta" or local_config != "":
                    config = linter.config.__dict__
                upload_to_server(
                    errors=errs,
                    paths=f_paths,
                    config=config,
                    url=linter.config.pyta_server_address,
                    version=__version__,
                )
        linter.generate_reports()
        return current_reporter
    except Exception as e:
        print(
            "[ERROR] Unexpected error encountered! Please report this to your instructor (and attach the code that caused the error)."
        )
        print('[ERROR] Error message: "{}"'.format(e))
        raise e


def _find_local_config(curr_dir):
    """Search for a `.pylintrc` configuration file provided in same (user)
    location as the source file to check.
    Return absolute path to the file, or None.
    `curr_dir` is an absolute path to a directory, containing a file to check.
    For more info see, pylint.config.find_pylintrc
    """
    if curr_dir.endswith(".py"):
        curr_dir = os.path.dirname(curr_dir)
    if os.path.exists(os.path.join(curr_dir, ".pylintrc")):
        return os.path.join(curr_dir, ".pylintrc")
    elif os.path.exists(os.path.join(curr_dir, "pylintrc")):
        return os.path.join(curr_dir, "pylintrc")


def _load_config(linter, config_location):
    """Load configuration into the linter."""
    linter.read_config_file(config_location)
    linter.config_file = config_location
    linter.load_config_file()


def reset_linter(config=None, file_linted=None):
    """Construct a new linter. Register config and checker plugins.

    To determine which configuration to use:
    - If the config argument is a string, use the config found at that location,
    - Otherwise,
        - Try to use the config file at directory of the file being linted,
        - Otherwise try to use default config file shipped with python_ta.
        - If the config argument is a dictionary, apply those options afterward.
    Do not re-use a linter object. Returns a new linter.
    """
    # Tuple of custom options. Note: 'type' must map to a value equal a key in the pylint/config/option.py `VALIDATORS` dict.
    new_checker_options = (
        (
            "pyta-type-check",
            {"default": False, "type": "yn", "metavar": "<yn>", "help": "Enable the type-checker."},
        ),
        (
            "pyta-number-of-messages",
            {
                "default": 5,
                "type": "int",
                "metavar": "<number_messages>",
                "help": "Display a certain number of messages to the user, without overwhelming them.",
            },
        ),
        (
            "pyta-template-file",
            {
                "default": "template.html.jinja",
                "type": "string",
                "metavar": "<pyta_reporter>",
                "help": "Template file for html format of htmlreporter output.",
            },
        ),
        (
            "pyta-error-permission",
            {
                "default": False,
                "type": "yn",
                "metavar": "<yn>",
                "help": "Permission to anonymously submit errors",
            },
        ),
        (
            "pyta-file-permission",
            {
                "default": False,
                "type": "yn",
                "metavar": "<yn>",
                "help": "Permission to anonymously submit files and errors",
            },
        ),
        (
            "pyta-server-address",
            {
                "default": "http://127.0.0.1:5000",
                "type": "string",
                "metavar": "<server-url>",
                "help": "Server address to submit anonymous data",
            },
        ),
    )

    custom_checkers = [
        "python_ta.checkers.forbidden_import_checker",
        "python_ta.checkers.possibly_undefined_checker",
        "python_ta.checkers.global_variables_checker",
        "python_ta.checkers.forbidden_io_function_checker",
        "python_ta.checkers.invalid_range_index_checker",
        "python_ta.checkers.one_iteration_checker",
        "python_ta.checkers.type_annotation_checker",
        "python_ta.checkers.unnecessary_indexing_checker",
        "python_ta.checkers.shadowing_in_comprehension_checker",
        "python_ta.checkers.redundant_assignment_checker",
        "python_ta.checkers.invalid_for_target_checker",
        "python_ta.checkers.missing_space_in_doctest_checker",
        "python_ta.checkers.pycodestyle_checker",
    ]

    # Register new options to a checker here to allow references to
    # options in `.pylintrc` config file.
    # Options stored in linter: `linter._all_options`, `linter._external_opts`
    linter = pylint.lint.PyLinter(options=new_checker_options)
    linter.load_default_plugins()  # Load checkers, reporters
    linter.load_plugin_modules(custom_checkers)

    if isinstance(config, str) and config != "":
        # Use config file at the specified path instead of the default.
        _load_config(linter, config)
    else:
        # If available, use config file at directory of the file being linted.
        pylintrc_location = None
        if file_linted:
            pylintrc_location = _find_local_config(file_linted)

        # Otherwise, use default config file shipped with python_ta package.
        if not pylintrc_location:
            pylintrc_location = _find_local_config(os.path.dirname(__file__))

        _load_config(linter, pylintrc_location)

        # Override part of the default config, with a dict of config options.
        # Note: these configs are overridden by config file in user's codebase
        # location.
        if isinstance(config, dict):
            for key in config:
                linter.global_set_option(key, config[key])

    # Custom checker configuration.
    if linter.config.pyta_type_check:
        linter.load_plugin_modules(["python_ta.checkers.type_inference_checker"])

    return linter


def get_file_paths(rel_path):
    """A generator for iterating python files within a directory.
    `rel_path` is a relative path to a file or directory.
    Returns paths to all files in a directory.
    """
    if not os.path.isdir(rel_path):
        yield rel_path  # Don't do anything; return the file name.
    else:
        for root, _, files in os.walk(rel_path):
            for filename in (f for f in files if f.endswith(".py")):
                yield os.path.join(root, filename)  # Format path, from root.


def _verify_pre_check(filepath):
    """Check student code for certain issues."""
    # Make sure the program doesn't crash for students.
    # Could use some improvement for better logging and error reporting.
    try:
        # Check for inline "pylint:" comment, which may indicate a student
        # trying to disable a check.
        with tokenize.open(os.path.expanduser(filepath)) as f:
            for tok_type, content, _, _, _ in tokenize.generate_tokens(f.readline):
                if tok_type != tokenize.COMMENT:
                    continue
                match = OPTION_PO.search(content)
                if match is not None:
                    print(
                        '[ERROR] String "pylint:" found in comment. '
                        + "No check run on file `{}.`\n".format(filepath)
                    )
                    return False
    except IndentationError as e:
        print(
            "[ERROR] python_ta could not check your code due to an "
            + "indentation error at line {}.".format(e.lineno)
        )
        return False
    except tokenize.TokenError as e:
        print(
            "[ERROR] python_ta could not check your code due to a " + "syntax error in your file."
        )
        return False
    except UnicodeDecodeError:
        print(
            "[ERROR] python_ta could not check your code due to an "
            + "invalid character. Please check the following lines "
            "in your file and all characters that are marked with a �."
        )
        with open(os.path.expanduser(filepath), encoding="utf-8", errors="replace") as f:
            for i, line in enumerate(f):
                if "�" in line:
                    print(f"  Line {i}: {line}", end="")
        return False
    return True


def _get_valid_files_to_check(module_name: str) -> Generator[str, None, None]:
    """A generator for all valid files to check."""
    # Allow call to check with empty args
    if module_name == "":
        m = sys.modules["__main__"]
        spec = importlib.util.spec_from_file_location(m.__name__, m.__file__)
        module_name = [spec.origin]
    # Enforce API to expect 1 file or directory if type is list
    elif isinstance(module_name, str):
        module_name = [module_name]
    # Otherwise, enforce API to expect `module_name` type as list
    elif not isinstance(module_name, list):
        print(
            "No checks run. Input to check, `{}`, has invalid type, must be a list of strings.".format(
                module_name
            )
        )
        return

    # Filter valid files to check
    for item in module_name:
        if not isinstance(item, str):  # Issue errors for invalid types
            print("No check run on file `{}`, with invalid type. Must be type: str.\n".format(item))
        elif os.path.isdir(item):
            yield item
        elif not os.path.exists(os.path.expanduser(item)):
            try:
                # For files with dot notation, e.g., `examples.<filename>`
                filepath = modutils.file_from_modpath(item.split("."))
                if os.path.exists(filepath):
                    yield filepath
                else:
                    print("Could not find the file called, `{}`\n".format(item))
            except ImportError:
                print("Could not find the file called, `{}`\n".format(item))
        else:
            yield item  # Check other valid files.


def doc(msg_id):
    """Open a webpage explaining the error for the given message."""
    msg_url = HELP_URL + "#" + msg_id
    print("Opening {} in a browser.".format(msg_url))
    webbrowser.open(msg_url)



from astroid import nodes
from hypothesis import HealthCheck, assume, given, settings

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


@given(cs.const_node())
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_const(node):
    """Test Const nodes representing int, bool, float, and None literal values.

    NOTE: string literals aren't checked here because it seems that astroid doesn't
    parse modules that contain only a single string literal.
    """
    assume(not isinstance(node.value, str))
    module, _ = cs._parse_text(node)
    cs._verify_type_setting(module, nodes.Const, type(node.value))
